@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2014-2017 INRIA
+  (C) 2014-2018 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -39,8 +39,13 @@
 #include "taler-merchant-httpd_pay.h"
 #include "taler-merchant-httpd_track-transaction.h"
 #include "taler-merchant-httpd_track-transfer.h"
+#include "taler-merchant-httpd_tip-authorize.h"
+#include "taler-merchant-httpd_tip-pickup.h"
+#include "taler-merchant-httpd_tip-query.h"
 #include "taler-merchant-httpd_history.h"
-#include "taler-merchant-httpd_map.h"
+#include "taler-merchant-httpd_refund.h"
+#include "taler-merchant-httpd_check-payment.h"
+#include "taler-merchant-httpd_trigger-pay.h"
 
 /**
  * Backlog for listen operation on unix-domain sockets.
@@ -72,12 +77,31 @@ static long long unsigned port;
  */
 struct GNUNET_TIME_Relative wire_transfer_delay;
 
+/**
+ * Locations from the configuration.  Mapping from
+ * label to location data.
+ */
+json_t *default_locations;
+
+/**
+ * If the frontend does NOT specify a payment deadline, how long should
+ * offers we make be valid by default?
+ */
+struct GNUNET_TIME_Relative default_pay_deadline;
+
 /**
  * Default maximum wire fee to assume, unless stated differently in the proposal
  * already.
  */
 struct TALER_Amount default_max_wire_fee;
 
+/**
+ * Default max deposit fee that the merchant is willing to
+ * pay; if deposit costs more, then the customer will cover
+ * the difference.
+ */
+struct TALER_Amount default_max_deposit_fee;
+
 /**
  * Default factor for wire fee amortization.
  */
@@ -175,13 +199,10 @@ url_handler (void *cls,
       { "/", MHD_HTTP_METHOD_GET, "text/plain",
         "Hello, I'm a merchant's Taler backend. This HTTP server is not for humans.\n", 0,
         &TMH_MHD_handler_static_response, MHD_HTTP_OK },
-      { "/proposal", MHD_HTTP_METHOD_POST, "application/json",
-        NULL, 0,
-        &MH_handler_proposal_put, MHD_HTTP_OK },
-      { "/pay", MHD_HTTP_METHOD_POST, "application/json",
+      { "/public/pay", MHD_HTTP_METHOD_POST, "application/json",
         NULL, 0,
         &MH_handler_pay, MHD_HTTP_OK },
-      { "/pay", NULL, "text/plain",
+      { "/public/pay", NULL, "text/plain",
         "Only POST is allowed", 0,
         &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED },
       { "/track/transfer", MHD_HTTP_METHOD_GET, "application/json",
@@ -199,13 +220,49 @@ url_handler (void *cls,
       { "/history", MHD_HTTP_METHOD_GET, "text/plain",
         "Only GET is allowed", 0,
         &MH_handler_history, MHD_HTTP_OK},
-      { "/proposal", MHD_HTTP_METHOD_GET, "text/plain",
-        "Only GET is allowed", 0,
+      { "/order", MHD_HTTP_METHOD_POST, "application/json",
+        NULL, 0,
+        &MH_handler_proposal_put, MHD_HTTP_OK },
+      { "/public/proposal", MHD_HTTP_METHOD_GET, "text/plain",
+        NULL, 0,
         &MH_handler_proposal_lookup, MHD_HTTP_OK},
       { "/proposal", NULL, "text/plain",
         "Only GET/POST are allowed", 0,
         &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED },
-
+      { "/refund", MHD_HTTP_METHOD_POST, "application/json",
+        NULL, 0,
+        &MH_handler_refund_increase, MHD_HTTP_OK},
+      { "/public/refund", MHD_HTTP_METHOD_GET, "text/plain",
+        NULL, 0,
+        &MH_handler_refund_lookup, MHD_HTTP_OK},
+      { "/refund", NULL, "application/json",
+        "Only POST/GET are allowed", 0,
+        &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED},
+      { "/tip-authorize", MHD_HTTP_METHOD_POST, "text/plain",
+        NULL, 0,
+        &MH_handler_tip_authorize, MHD_HTTP_OK},
+      { "/tip-authorize", NULL, "application/json",
+        "Only POST is allowed", 0,
+        &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED},
+      /* backwards compatibility alias for /public/tip-pickup */
+      { "/tip-pickup", MHD_HTTP_METHOD_POST, "text/plain",
+        NULL, 0,
+        &MH_handler_tip_pickup, MHD_HTTP_OK},
+      { "/public/tip-pickup", MHD_HTTP_METHOD_POST, "text/plain",
+        NULL, 0,
+        &MH_handler_tip_pickup, MHD_HTTP_OK},
+      { "/tip-pickup", NULL, "application/json",
+        "Only POST is allowed", 0,
+        &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED},
+      { "/tip-query", MHD_HTTP_METHOD_GET, "text/plain",
+        NULL, 0,
+        &MH_handler_tip_query, MHD_HTTP_OK},
+      { "/check-payment", MHD_HTTP_METHOD_GET, "text/plain",
+        NULL, 0,
+        &MH_handler_check_payment, MHD_HTTP_OK},
+      { "/public/trigger-pay", MHD_HTTP_METHOD_GET, "text/plain",
+        NULL, 0,
+        &MH_handler_trigger_pay, MHD_HTTP_OK},
       {NULL, NULL, NULL, NULL, 0, 0 }
     };
   static struct TMH_RequestHandler h404 =
@@ -214,28 +271,28 @@ url_handler (void *cls,
       "<html><title>404: not found</title></html>", 0,
       &TMH_MHD_handler_static_response, MHD_HTTP_NOT_FOUND
     };
-  struct TM_HandlerContext *hc;
-  struct TMH_RequestHandler *rh;
-  unsigned int i;
-  int ret;
-
   GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-              "Handling request for URL `%s'\n",
+              "Handling request (%s) for URL `%s'\n",
+              method,
               url);
-  for (i=0;NULL != handlers[i].url;i++)
+  for (unsigned int i=0;NULL != handlers[i].url;i++)
   {
-    rh = &handlers[i];
+    struct TMH_RequestHandler *rh = &handlers[i];
+
     if ( (0 == strcasecmp (url,
                            rh->url)) &&
          ( (NULL == rh->method) ||
            (0 == strcasecmp (method,
                              rh->method)) ) )
     {
+      struct TM_HandlerContext *hc;
+      int ret;
+
       ret = rh->handler (rh,
-                         connection,
-                         con_cls,
-                         upload_data,
-                         upload_data_size);
+			 connection,
+			 con_cls,
+			 upload_data,
+			 upload_data_size);
       hc = *con_cls;
       if (NULL != hc)
         hc->rh = rh;
@@ -263,10 +320,21 @@ hashmap_free (void *cls,
               void *value)
 {
   struct MerchantInstance *mi = value;
+  struct WireMethod *wm;
+
+  while (NULL != (wm = (mi->wm_head)))
+  {
+    GNUNET_CONTAINER_DLL_remove (mi->wm_head,
+                                 mi->wm_tail,
+                                 wm);
+    json_decref (wm->j_wire);
+    GNUNET_free (wm);
+  }
 
-  json_decref (mi->j_wire);
   GNUNET_free (mi->id);
   GNUNET_free (mi->keyfile);
+  GNUNET_free (mi->name);
+  GNUNET_free_non_null (mi->tip_exchange);
   GNUNET_free (mi);
   return GNUNET_YES;
 }
@@ -340,7 +408,7 @@ handle_mhd_completion_callback (void *cls,
 
   if (NULL == hc)
     return;
-  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
               "Finished handling request for `%s' with status %d\n",
               hc->rh->url,
               (int) toe);
@@ -357,6 +425,12 @@ static struct GNUNET_SCHEDULER_Task *
 prepare_daemon (void);
 
 
+/**
+ * Set if we should immediately #MHD_run again.
+ */
+static int triggered;
+
+
 /**
  * Call MHD to process pending requests and then go back
  * and schedule the next run.
@@ -367,7 +441,10 @@ static void
 run_daemon (void *cls)
 {
   mhd_task = NULL;
-  GNUNET_assert (MHD_YES == MHD_run (mhd));
+  do {
+    triggered = 0;
+    GNUNET_assert (MHD_YES == MHD_run (mhd));
+  } while (0 != triggered);
   mhd_task = prepare_daemon ();
 }
 
@@ -381,8 +458,16 @@ run_daemon (void *cls)
 void
 TMH_trigger_daemon ()
 {
-  GNUNET_SCHEDULER_cancel (mhd_task);
-  run_daemon (NULL);
+  if (NULL != mhd_task)
+  {
+    GNUNET_SCHEDULER_cancel (mhd_task);
+    mhd_task = NULL;
+    run_daemon (NULL);
+  }
+  else
+  {
+    triggered = 1;
+  }
 }
 
 
@@ -427,42 +512,200 @@ prepare_daemon ()
   GNUNET_NETWORK_fdset_copy_native (wws, &ws, max + 1);
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
               "Adding run_daemon select task\n");
-  ret =
-      GNUNET_SCHEDULER_add_select (GNUNET_SCHEDULER_PRIORITY_HIGH,
-				   tv, wrs, wws,
-                                   &run_daemon,
-                                   NULL);
+  ret = GNUNET_SCHEDULER_add_select (GNUNET_SCHEDULER_PRIORITY_HIGH,
+				     tv,
+				     wrs,
+				     wws,
+				     &run_daemon,
+				     NULL);
   GNUNET_NETWORK_fdset_destroy (wrs);
   GNUNET_NETWORK_fdset_destroy (wws);
   return ret;
 }
 
 
+/**
+ * Callback that looks for 'merchant-location-*' sections,
+ * and populates @a default_locations.
+ *
+ * @param cls closure
+ * @section section name this callback gets
+ */
+static void
+locations_iterator_cb (void *cls,
+                       const char *section)
+{
+  static const char *keys[] = {
+    "country",
+    "city",
+    "state",
+    "region",
+    "province",
+    "zip_code",
+    "street",
+    "street_number",
+    NULL,
+  };
+  struct GNUNET_CONFIGURATION_Handle *cfg = cls;
+  const char *prefix = "merchant-location-";
+  const char *substr = strstr (section, prefix);
+  const char *locname;
+  json_t *loc;
+
+  if ( (NULL == substr) || (substr != section) )
+    return;
+  locname = section + strlen (prefix);
+  if (0 == strlen (locname))
+    return;
+  GNUNET_assert (json_is_object (default_locations));
+
+  loc = json_object ();
+  json_object_set_new (default_locations,
+                       locname,
+                       loc);
+  for (unsigned int pos = 0; NULL != keys[pos]; pos++)
+  {
+    char *val;
+
+    if (GNUNET_OK ==
+        GNUNET_CONFIGURATION_get_value_string (cfg,
+                                               section,
+                                               keys[pos],
+                                               &val))
+    {
+      json_object_set_new (loc,
+                           keys[pos],
+                           json_string (val));
+      GNUNET_free (val);
+    }
+  }
+}
+
+
+/**
+ * Closure for the #wireformat_iterator_cb().
+ */
+struct WireFormatIteratorContext
+{
+  /**
+   * The global iteration context.
+   */
+  struct IterateInstancesCls *iic;
+
+  /**
+   * The merchant instance we are currently building.
+   */
+  struct MerchantInstance *mi;
+};
+
+
+/**
+ * Callback that looks for 'merchant-instance-wireformat-*' sections,
+ * and populates our wire method according to the data
+ *
+ * @param cls closure with a `struct WireFormatIteratorContext *`
+ * @section section name this callback gets
+ */
+static void
+wireformat_iterator_cb (void *cls,
+                        const char *section)
+{
+  struct WireFormatIteratorContext *wfic = cls;
+  struct MerchantInstance *mi = wfic->mi;
+  struct IterateInstancesCls *iic = wfic->iic;
+  char *instance_wiresection;
+  struct WireMethod *wm;
+  json_t *type;
+  char *emsg;
+
+  GNUNET_asprintf (&instance_wiresection,
+                   "merchant-instance-wireformat-%s",
+                   mi->id);
+  if (0 != strncmp (section,
+                    instance_wiresection,
+                    strlen (instance_wiresection)))
+  {
+    GNUNET_free (instance_wiresection);
+    return;
+  }
+  GNUNET_free (instance_wiresection);
+
+  wm = GNUNET_new (struct WireMethod);
+  /* FIXME: maybe use sorting to address #4939-12806? */
+  GNUNET_CONTAINER_DLL_insert (mi->wm_head,
+                               mi->wm_tail,
+                               wm);
+  wm->j_wire = iic->plugin->get_wire_details (iic->plugin->cls,
+                                              iic->config,
+                                              section);
+  if ( (NULL == (type = json_object_get (wm->j_wire,
+                                         "type"))) ||
+       (! json_is_string (type)) )
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Malformed wireformat: lacks type\n");
+    iic->ret |= GNUNET_SYSERR;
+    return;
+  }
+  wm->wire_method = json_string_value (type);
+
+  if (TALER_EC_NONE !=
+      iic->plugin->wire_validate (iic->plugin->cls,
+                                  wm->j_wire,
+                                  NULL,
+                                  &emsg))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Malformed wireformat: %s\n",
+                emsg);
+    GNUNET_free (emsg);
+    iic->ret |= GNUNET_SYSERR;
+    return;
+  }
+
+  if (GNUNET_OK !=
+      TALER_JSON_hash (wm->j_wire,
+                       &wm->h_wire))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to hash wireformat\n");
+    iic->ret |= GNUNET_SYSERR;
+  }
+
+#define EXTRADEBUG
+#ifdef EXTRADEBUGG
+  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+              "Found wireformat instance:\n");
+  json_dumpf (wm->j_wire,
+              stdout,
+              0);
+  printf ("\n");
+#endif
+}
+
+
 /**
  * Callback that looks for 'merchant-instance-*' sections,
  * and populates accordingly each instance's data
  *
- * @param cls closure
+ * @param cls closure of type `struct IterateInstancesCls`
  * @section section name this callback gets
  */
 static void
 instances_iterator_cb (void *cls,
                        const char *section)
 {
-  char *substr;
+  struct IterateInstancesCls *iic = cls;
   char *token;
-  char *instance_wiresection;
   struct MerchantInstance *mi;
-  struct IterateInstancesCls *iic;
   struct GNUNET_CRYPTO_EddsaPrivateKey *pk;
   /* used as hashmap keys */
   struct GNUNET_HashCode h_pk;
   struct GNUNET_HashCode h_id;
-  json_t *type;
-  char *emsg;
+  const char *substr;
 
-  iic = cls;
-  substr = strstr (section, "merchant-instance-");
+  substr = strstr (section,
+                   "merchant-instance-");
 
   if ( (NULL == substr) ||
        (NULL != strstr (section,
@@ -483,6 +726,19 @@ instances_iterator_cb (void *cls,
               "Extracted token: %s\n",
               token + 1);
   mi = GNUNET_new (struct MerchantInstance);
+  if (GNUNET_OK !=
+      GNUNET_CONFIGURATION_get_value_string (iic->config,
+                                             section,
+                                             "NAME",
+                                             &mi->name))
+  {
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               section,
+                               "NAME");
+    GNUNET_free (mi);
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
 
   if (GNUNET_OK !=
       GNUNET_CONFIGURATION_get_value_filename (iic->config,
@@ -493,12 +749,56 @@ instances_iterator_cb (void *cls,
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                section,
                                "KEYFILE");
+    GNUNET_free (mi->name);
     GNUNET_free (mi);
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
+  if (GNUNET_OK ==
+      GNUNET_CONFIGURATION_get_value_string (iic->config,
+                                             section,
+                                             "TIP_EXCHANGE",
+                                             &mi->tip_exchange))
+  {
+    char *tip_reserves;
+    struct GNUNET_CRYPTO_EddsaPrivateKey *pk;
 
-  if (GNUNET_YES != GNUNET_DISK_file_test (mi->keyfile))
+    if (GNUNET_OK !=
+        GNUNET_CONFIGURATION_get_value_filename (iic->config,
+                                                 section,
+                                                 "TIP_RESERVE_PRIV_FILENAME",
+                                                 &tip_reserves))
+    {
+      GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                                 section,
+                                 "TIP_RESERVE_PRIV_FILENAME");
+      GNUNET_free (mi->keyfile);
+      GNUNET_free (mi->name);
+      GNUNET_free (mi);
+      GNUNET_SCHEDULER_shutdown ();
+      return;
+    }
+    pk = GNUNET_CRYPTO_eddsa_key_create_from_file (tip_reserves);
+    if (NULL == pk)
+    {
+      GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
+                                 section,
+                                 "TIP_RESERVE_PRIV_FILENAME",
+                                 "Failed to read private key");
+      GNUNET_free (tip_reserves);
+      GNUNET_free (mi->keyfile);
+      GNUNET_free (mi->name);
+      GNUNET_free (mi);
+      GNUNET_SCHEDULER_shutdown ();
+      return;
+    }
+    mi->tip_reserve.eddsa_priv = *pk;
+    GNUNET_free (pk);
+    GNUNET_free (tip_reserves);
+  }
+
+  if (GNUNET_YES !=
+      GNUNET_DISK_file_test (mi->keyfile))
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                 "Merchant private key `%s' does not exist yet, creating it!\n",
                 mi->keyfile);
@@ -507,6 +807,7 @@ instances_iterator_cb (void *cls,
   {
     GNUNET_break (0);
     GNUNET_free (mi->keyfile);
+    GNUNET_free (mi->name);
     GNUNET_free (mi);
     GNUNET_SCHEDULER_shutdown ();
     return;
@@ -517,60 +818,33 @@ instances_iterator_cb (void *cls,
   GNUNET_free (pk);
 
   mi->id = GNUNET_strdup (token + 1);
-  if (0 == strcmp ("default", mi->id))
+  if (0 == strcasecmp ("default",
+                       mi->id))
     iic->default_instance = GNUNET_YES;
 
-  GNUNET_asprintf (&instance_wiresection,
-                   "merchant-instance-wireformat-%s",
-                   mi->id);
-  mi->j_wire = iic->plugin->get_wire_details (iic->plugin->cls,
-                                              iic->config,
-                                              instance_wiresection);
-  GNUNET_free (instance_wiresection);
-  if ( (NULL == (type = json_object_get (mi->j_wire,
-                                         "type"))) ||
-       (! json_is_string (type)) )
+  /* Initialize wireformats */
   {
-    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                "Malformed wireformat: lacks type\n");
-    iic->ret |= GNUNET_SYSERR;
-  }
-  mi->wire_method = json_string_value (type);
+    struct WireFormatIteratorContext wfic = {
+      .iic = iic,
+      .mi = mi
+    };
 
-  if (TALER_EC_NONE !=
-      iic->plugin->wire_validate (iic->plugin->cls,
-                                  mi->j_wire,
-                                  NULL,
-                                  &emsg))
-  {
-    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                "Malformed wireformat: %s\n",
-                emsg);
-    GNUNET_free (emsg);
-    iic->ret |= GNUNET_SYSERR;
+    GNUNET_CONFIGURATION_iterate_sections (iic->config,
+                                           &wireformat_iterator_cb,
+                                           &wfic);
   }
 
-  if (GNUNET_YES != TALER_JSON_hash (mi->j_wire,
-                                     &mi->h_wire))
+  if (NULL == mi->wm_head)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                "Failed to hash wireformat\n");
+                "Failed to load wire formats for instance `%s'\n",
+                mi->id);
     iic->ret |= GNUNET_SYSERR;
   }
-  #define EXTRADEBUG
-  #ifdef EXTRADEBUGG
-  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
-              "Found wireformat instance:\n");
-              json_dumpf (mi->j_wire, stdout, 0);
-              printf ("\n");
-  #endif
 
   GNUNET_CRYPTO_hash (mi->id,
                       strlen (mi->id),
                       &h_id);
-  GNUNET_CRYPTO_hash (&mi->pubkey.eddsa_pub,
-                      sizeof (struct GNUNET_CRYPTO_EddsaPublicKey),
-                      &h_pk);
   if (GNUNET_OK !=
       GNUNET_CONTAINER_multihashmap_put (by_id_map,
                                          &h_id,
@@ -581,6 +855,9 @@ instances_iterator_cb (void *cls,
                 "Failed to put an entry into the 'by_id' hashmap\n");
     iic->ret |= GNUNET_SYSERR;
   }
+  GNUNET_CRYPTO_hash (&mi->pubkey.eddsa_pub,
+                      sizeof (struct GNUNET_CRYPTO_EddsaPublicKey),
+                      &h_pk);
   if (GNUNET_OK !=
       GNUNET_CONTAINER_multihashmap_put (by_kpub_map,
                                          &h_pk,
@@ -633,7 +910,7 @@ TMH_lookup_instance (const char *name)
  * a wrong instance
  */
 struct MerchantInstance *
-get_instance (struct json_t *json)
+TMH_lookup_instance_json (struct json_t *json)
 {
   struct json_t *instance;
   const char *instance_str;
@@ -646,6 +923,23 @@ get_instance (struct json_t *json)
 }
 
 
+/**
+ * Iterate over locations in config in order to populate
+ * the location data.
+ *
+ * @param config configuration handle
+ * @return #GNUNET_OK if successful, #GNUNET_SYSERR upon errors
+ */
+static void
+iterate_locations (const struct GNUNET_CONFIGURATION_Handle *config)
+{
+  GNUNET_assert (NULL == default_locations);
+  default_locations = json_object ();
+  GNUNET_CONFIGURATION_iterate_sections (config,
+                                         &locations_iterator_cb,
+                                         (void *) config);
+}
+
 /**
  * Iterate over each merchant instance, in order to populate
  * each instance's own data
@@ -655,7 +949,7 @@ get_instance (struct json_t *json)
  * @return #GNUNET_OK if successful, #GNUNET_SYSERR upon errors
  * (for example, if no "default" instance is defined)
  */
-static unsigned int
+static int
 iterate_instances (const struct GNUNET_CONFIGURATION_Handle *config,
                    const char *allowed)
 {
@@ -682,7 +976,7 @@ iterate_instances (const struct GNUNET_CONFIGURATION_Handle *config,
   iic->plugin->library_name = lib_name;
   GNUNET_CONFIGURATION_iterate_sections (config,
                                          &instances_iterator_cb,
-                                         (void *) iic);
+                                         iic);
 
   if (GNUNET_NO == iic->default_instance)
   {
@@ -708,14 +1002,11 @@ iterate_instances (const struct GNUNET_CONFIGURATION_Handle *config,
   return GNUNET_OK;
 
  fail:
-  do {
-    GNUNET_PLUGIN_unload (lib_name,
-                          iic->plugin);
-    GNUNET_free (lib_name);
-    GNUNET_free (iic);
-    GNUNET_SCHEDULER_shutdown ();
-    return GNUNET_SYSERR;
-  } while (0);
+  GNUNET_PLUGIN_unload (lib_name,
+			iic->plugin);
+  GNUNET_free (lib_name);
+  GNUNET_free (iic);
+  return GNUNET_SYSERR;
 }
 
 
@@ -737,12 +1028,14 @@ run (void *cls,
   char *wireformat;
   int fh;
 
+  GNUNET_log (GNUNET_ERROR_TYPE_INFO, "running taler-merchant-httpd\n");
+
   result = GNUNET_SYSERR;
   GNUNET_SCHEDULER_add_shutdown (&do_shutdown,
                                  NULL);
   GNUNET_assert (GNUNET_OK ==
                  GNUNET_log_setup ("taler-merchant-httpd",
-                                   "DEBUG",
+                                   "WARNING",
                                    NULL));
   if (GNUNET_SYSERR ==
       TMH_EXCHANGES_init (config))
@@ -784,39 +1077,43 @@ run (void *cls,
     return;
   }
 
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_time (config,
+                                           "merchant",
+                                           "DEFAULT_PAY_DEADLINE",
+                                           &default_pay_deadline))
+  {
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               "merchant",
+                               "DEFAULT_PAY_DEADLINE");
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+
   if (GNUNET_OK !=
       TALER_config_get_denom (config,
                               "merchant",
                               "DEFAULT_MAX_WIRE_FEE",
                               &default_max_wire_fee))
   {
-    char *currency;
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               "merchant",
+                               "DEFAULT_MAX_WIRE_FEE");
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
 
-    if (GNUNET_OK !=
-        GNUNET_CONFIGURATION_get_value_string (config,
-                                               "taler",
-                                               "CURRENCY",
-                                               &currency))
-    {
-      GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
-                                 "taler",
-                                 "CURRENCY");
-      GNUNET_SCHEDULER_shutdown ();
-      return;
-    }
-    if (GNUNET_OK !=
-        TALER_amount_get_zero (currency,
-                               &default_max_wire_fee))
-    {
-      GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
-                                 "taler",
-                                 "CURRENCY",
-                                 "Specified value not legal for a Taler currency");
-      GNUNET_SCHEDULER_shutdown ();
-      GNUNET_free (currency);
-      return;
-    }
-    GNUNET_free (currency);
+  if (GNUNET_OK !=
+      TALER_config_get_denom (config,
+                              "merchant",
+                              "DEFAULT_MAX_DEPOSIT_FEE",
+                              &default_max_deposit_fee))
+  {
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               "merchant",
+                               "DEFAULT_MAX_DEPOSIT_FEE");
+    GNUNET_SCHEDULER_shutdown ();
+    return;
   }
 
   if (GNUNET_OK !=
@@ -845,10 +1142,18 @@ run (void *cls,
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
-  iterate_instances (config,
-                     wireformat);
+  if (GNUNET_OK !=
+      iterate_instances (config,
+			 wireformat))
+  {
+    GNUNET_free (wireformat);
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
   GNUNET_free (wireformat);
 
+  iterate_locations (config);
+
   if (NULL ==
       (db = TALER_MERCHANTDB_plugin_load (config)))
   {
@@ -1124,7 +1429,8 @@ run (void *cls,
  * @return 0 ok, 1 on error
  */
 int
-main (int argc, char *const *argv)
+main (int argc,
+      char *const *argv)
 {
   struct GNUNET_GETOPT_CommandLineOption options[] = {
 
