@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015 INRIA
+  Copyright (C) 2014-2018 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
@@ -72,7 +72,40 @@ struct IterateInstancesCls
 
 
 /**
- * Information that defines a merchant "instance". That way, a single
+ * Supported wire method.  Kept in a DLL.
+ */
+struct WireMethod
+{
+  /**
+   * Next entry in DLL.
+   */
+  struct WireMethod *next;
+
+  /**
+   * Previous entry in DLL.
+   */
+  struct WireMethod *prev;
+
+  /**
+   * Which wire method is @e j_wire using?  Points into @e j_wire.
+   */
+  const char *wire_method;
+
+  /**
+   * Wire details for this instance
+   */
+  struct json_t *j_wire;
+
+  /**
+   * Hash of our wire format details as given in #j_wire.
+   */
+  struct GNUNET_HashCode h_wire;
+
+};
+
+
+/**
+ * Information that defines a merchant "instance". Tha4673t way, a single
  * backend can account for several merchants, as used to do in donation
  * shops
  */
@@ -86,28 +119,25 @@ struct MerchantInstance
    */
   char *id;
 
+  /**
+   * Legal name of the merchant.
+   */
+  char *name;
+
   /**
    * File holding the merchant's private key
    */
   char *keyfile;
 
-  /* NOTE: the *_wire-fields should eventually be moved into a DLL
-     once we implement #4939 */
+  /**
+   * Next entry in DLL.
+   */
+  struct WireMethod *wm_head;
 
   /**
-   * Which wire method is @e j_wire using?
+   * Previous entry in DLL.
    */
-  const char *wire_method;
-
-  /**
-   * Wire details for this instance
-   */
-  struct json_t *j_wire;
-
-  /**
-   * Hash of our wire format details as given in #j_wire.
-   */
-  struct GNUNET_HashCode h_wire;
+  struct WireMethod *wm_tail;
 
   /**
    * Merchant's private key
@@ -119,6 +149,17 @@ struct MerchantInstance
    */
   struct TALER_MerchantPublicKeyP pubkey;
 
+  /**
+   * Exchange this instance uses for tipping, NULL if tipping
+   * is not supported.
+   */
+  char *tip_exchange;
+
+  /**
+   * What is the private key of the reserve used for signing tips by this exchange?
+   * Only valid if @e tip_exchange is non-null.
+   */
+  struct TALER_ReservePrivateKeyP tip_reserve;
 };
 
 
@@ -220,9 +261,10 @@ struct TM_HandlerContext
 
 
 /**
- * Our wire format details in JSON format (with salt).
+ * Locations from the configuration.  Mapping from
+ * label to location data.
  */
-extern json_t *j_wire;
+extern json_t *default_locations;
 
 /**
  * Default maximum wire fee to assume, unless stated differently in the proposal
@@ -230,6 +272,13 @@ extern json_t *j_wire;
  */
 extern struct TALER_Amount default_max_wire_fee;
 
+/**
+ * Default max deposit fee that the merchant is willing to
+ * pay; if deposit costs more, then the customer will cover
+ * the difference.
+ */
+extern struct TALER_Amount default_max_deposit_fee;
+
 /**
  * Default factor for wire fee amortization.
  */
@@ -268,6 +317,12 @@ extern struct TALER_MERCHANTDB_Plugin *db;
  */
 extern struct GNUNET_TIME_Relative wire_transfer_delay;
 
+/**
+ * If the frontend does NOT specify a payment deadline, how long should
+ * offers we make be valid by default?
+ */
+extern struct GNUNET_TIME_Relative default_pay_deadline;
+
 /**
  * Kick MHD to run now, to be called after MHD_resume_connection().
  * Basically, we need to explicitly resume MHD's event loop whenever
@@ -287,4 +342,21 @@ struct MerchantInstance *
 TMH_lookup_instance (const char *name);
 
 
+/**
+ * Extract merchant instance from the given JSON
+ *
+ * @param json the JSON to inspect; it is not required to
+ * comply with any particular format. It will only be checked
+ * if the field "instance" is there.
+ * @return a pointer to a #struct MerchantInstance. This will be
+ * the 'default' merchant if the frontend did not specif any
+ * "instance" field. The user should not care to free the returned
+ * value, as it is taken from a global array that will be freed
+ * by the general shutdown routine. NULL if the frontend specified
+ * a wrong instance
+ */
+struct MerchantInstance *
+TMH_lookup_instance_json (struct json_t *json);
+
+
 #endif
