@@ -153,9 +153,9 @@ struct Exchange
   struct TMH_EXCHANGES_FindOperation *fo_tail;
 
   /**
-   * (base) URI of the exchange.
+   * (base) URL of the exchange.
    */
-  char *uri;
+  char *url;
 
   /**
    * A connection to this exchange
@@ -255,10 +255,12 @@ json_t *trusted_exchanges;
  *   in order to get the "good" keys.
  * @param keys information about the various keys used
  *        by the exchange
+ * @param compat version compatibility data
  */
 static void
 keys_mgmt_cb (void *cls,
-              const struct TALER_EXCHANGE_Keys *keys);
+              const struct TALER_EXCHANGE_Keys *keys,
+	      enum TALER_EXCHANGE_VersionCompatibility compat);
 
 
 /**
@@ -277,14 +279,14 @@ retry_exchange (void *cls)
   exchange->retry_task = NULL;
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
               "Connecting to exchange exchange %s in retry_exchange\n",
-              exchange->uri);
+              exchange->url);
   if (NULL != exchange->conn)
   {
     TALER_EXCHANGE_disconnect (exchange->conn);
     exchange->conn = NULL;
   }
   exchange->conn = TALER_EXCHANGE_connect (merchant_curl_ctx,
-                                           exchange->uri,
+                                           exchange->url,
                                            &keys_mgmt_cb,
                                            exchange,
                                            TALER_EXCHANGE_OPTION_END);
@@ -310,7 +312,12 @@ process_wire_fees (void *cls,
   struct FeesByWireMethod *f;
   struct TALER_EXCHANGE_WireAggregateFees *endp;
   struct TALER_EXCHANGE_WireAggregateFees *af;
+  const struct TALER_EXCHANGE_Keys *keys;
+  const struct TALER_MasterPublicKeyP *master_pub;
 
+  keys = TALER_EXCHANGE_get_keys (exchange->conn);
+  GNUNET_assert (NULL != keys);
+  master_pub = &keys->master_pub;
   for (f = exchange->wire_fees_head; NULL != f; f = f->next)
     if (0 == strcasecmp (wire_method,
                          f->wire_method))
@@ -339,15 +346,74 @@ process_wire_fees (void *cls,
   }
   while (NULL != fees)
   {
+    struct GNUNET_HashCode h_wire_method;
+    enum GNUNET_DB_QueryStatus qs;
+
     af = GNUNET_new (struct TALER_EXCHANGE_WireAggregateFees);
     *af = *fees;
+    GNUNET_CRYPTO_hash (wire_method,
+			strlen (wire_method) + 1,
+			&h_wire_method);
+    GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+		"Storing wire fee for `%s' and method `%s' at %s in DB; the fee is %s\n",
+		TALER_B2S (master_pub),
+		wire_method,
+		GNUNET_STRINGS_absolute_time_to_string (af->start_date),
+		TALER_amount2s (&af->wire_fee));
+    db->preflight (db->cls);
+    if (GNUNET_OK !=
+        db->start (db->cls,
+                   "store wire fee"))
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+		  "Failed to start database transaction!\n");
+      GNUNET_free (af);
+      break;
+    }
+    qs = db->store_wire_fee_by_exchange (db->cls,
+					 master_pub,
+					 &h_wire_method,
+					 &af->wire_fee,
+					 &af->closing_fee,
+					 af->start_date,
+					 af->end_date,
+					 &af->master_sig);
+    if (0 > qs)
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+		  "Failed to persist exchange wire fees in merchant DB!\n");
+      GNUNET_free (af);
+      fees = fees->next;
+      db->rollback (db->cls);
+      continue;
+    }
+    if (0 == qs)
+    {
+      /* Entry was already in DB, fine, continue as if we had succeeded */
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+		  "Fees already in DB, rolling back transaction attempt!\n");
+      db->rollback (db->cls);
+    }
+    if (0 < qs)
+    {
+      /* Inserted into DB, make sure transaction completes */
+      qs = db->commit (db->cls);
+      if (0 > qs)
+      {
+        GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                    "Failed to persist exchange wire fees in merchant DB!\n");
+        GNUNET_free (af);
+        fees = fees->next;
+        continue;
+      }
+    }
+
     af->next = NULL;
     if (NULL == endp)
       f->af = af;
     else
       endp->next = af;
     endp = af;
-    // FIXME: also preserve `fees` in backend DB (under wire method + exchange master pub!)
     fees = fees->next;
   }
 }
@@ -398,14 +464,15 @@ get_wire_fees (struct Exchange *exchange,
 static int
 process_find_operations (struct Exchange *exchange)
 {
-  struct TMH_EXCHANGES_FindOperation *fo;
   struct TMH_EXCHANGES_FindOperation *fn;
   struct GNUNET_TIME_Absolute now;
   int need_wire;
 
   now = GNUNET_TIME_absolute_get ();
   need_wire = GNUNET_NO;
-  for (fo = exchange->fo_head; NULL != fo; fo = fn)
+  for (struct TMH_EXCHANGES_FindOperation *fo = exchange->fo_head;
+       NULL != fo;
+       fo = fn)
   {
     const struct TALER_Amount *wire_fee;
 
@@ -493,21 +560,24 @@ handle_wire_data (void *cls,
                   const json_t *obj)
 {
   struct Exchange *exchange = cls;
+  const struct TALER_EXCHANGE_Keys *keys;
 
   exchange->wire_request = NULL;
   if (MHD_HTTP_OK != http_status)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                 "Failed to obtain /wire details from `%s': %d\n",
-                exchange->uri,
+                exchange->url,
                 ec);
     return;
   }
-  if (GNUNET_OK !=
-      TALER_EXCHANGE_wire_get_fees (&TALER_EXCHANGE_get_keys (exchange->conn)->master_pub,
-                                    obj,
-                                    &process_wire_fees,
-                                    exchange))
+  keys = TALER_EXCHANGE_get_keys (exchange->conn);
+  if ( (NULL == keys) ||
+       (GNUNET_OK !=
+        TALER_EXCHANGE_wire_get_fees (&keys->master_pub,
+                                      obj,
+                                      &process_wire_fees,
+                                      exchange)) )
   {
     /* Report hard failure to all callbacks! */
     struct TMH_EXCHANGES_FindOperation *fo;
@@ -588,10 +658,12 @@ wire_task_cb (void *cls)
  *   in order to get the "good" keys.
  * @param keys information about the various keys used
  *        by the exchange
+ * @param compat version compatibility data
  */
 static void
 keys_mgmt_cb (void *cls,
-              const struct TALER_EXCHANGE_Keys *keys)
+              const struct TALER_EXCHANGE_Keys *keys,
+	      enum TALER_EXCHANGE_VersionCompatibility compat)
 {
   struct Exchange *exchange = cls;
   struct GNUNET_TIME_Absolute expire;
@@ -610,24 +682,49 @@ keys_mgmt_cb (void *cls,
       GNUNET_SCHEDULER_cancel (exchange->wire_task);
       exchange->wire_task = NULL;
     }
+    if (TALER_EXCHANGE_VC_INCOMPATIBLE_NEWER == compat)
+    {
+      /* Give up, log hard error. */
+      GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+		  "Exchange `%s' runs an incompatible more recent version of the Taler protocol. Will not retry. This client may need to be updated.\n",
+		  exchange->url);
+      return;
+    }
     exchange->retry_delay = RETRY_BACKOFF (exchange->retry_delay);
     GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                 "Failed to fetch /keys from `%s', retrying in %s\n",
-                exchange->uri,
+                exchange->url,
                 GNUNET_STRINGS_relative_time_to_string (exchange->retry_delay,
                                                         GNUNET_YES));
+    GNUNET_assert (NULL == exchange->retry_task);
     exchange->retry_task = GNUNET_SCHEDULER_add_delayed (exchange->retry_delay,
                                                          &retry_exchange,
                                                          exchange);
     return;
   }
-  expire = TALER_EXCHANGE_check_keys_current (exchange->conn);
+  if (0 != (TALER_EXCHANGE_VC_NEWER & compat))
+  {
+    /* Warn user exactly once about need to upgrade */
+    static int once;
+
+    if (0 == once)
+    {
+      once = 1;
+      GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+		  "Exchange `%s' runs a more recent version of the Taler protocol. You may want to update this client.\n",
+		  exchange->url);
+    }
+  }
+  expire = TALER_EXCHANGE_check_keys_current (exchange->conn,
+                                              GNUNET_NO);
   if (0 == expire.abs_value_us)
     delay = RELOAD_DELAY;
   else
     delay = GNUNET_TIME_absolute_get_remaining (expire);
   exchange->retry_delay
     = GNUNET_TIME_UNIT_ZERO;
+  if (NULL != exchange->retry_task)
+    GNUNET_SCHEDULER_cancel (exchange->retry_task);
   exchange->retry_task
     = GNUNET_SCHEDULER_add_delayed (delay,
                                     &retry_exchange,
@@ -680,7 +777,7 @@ return_result (void *cls)
  * to the exchange, or if it is not acceptable, @a fc is called with
  * NULL for the exchange.
  *
- * @param chosen_exchange URI of the exchange we would like to talk to
+ * @param chosen_exchange URL of the exchange we would like to talk to
  * @param wire_method the wire method we will use with @a chosen_exchange, NULL for none
  * @param fc function to call with the handles for the exchange
  * @param fc_cls closure for @a fc
@@ -694,6 +791,7 @@ TMH_EXCHANGES_find_exchange (const char *chosen_exchange,
 {
   struct Exchange *exchange;
   struct TMH_EXCHANGES_FindOperation *fo;
+  struct GNUNET_TIME_Absolute now;
 
   if (NULL == merchant_curl_ctx)
   {
@@ -708,9 +806,9 @@ TMH_EXCHANGES_find_exchange (const char *chosen_exchange,
   /* Check if the exchange is known */
   for (exchange = exchange_head; NULL != exchange; exchange = exchange->next)
     /* test it by checking public key --- FIXME: hostname or public key!?
-       Should probably be URI, not hostname anyway! */
+       Should probably be URL, not hostname anyway! */
   {
-    if (0 == strcmp (exchange->uri,
+    if (0 == strcmp (exchange->url,
                      chosen_exchange))
     {
       GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
@@ -720,13 +818,13 @@ TMH_EXCHANGES_find_exchange (const char *chosen_exchange,
     }
     GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
                 "Comparing chosen exchange url '%s' with known url '%s'.\n",
-                chosen_exchange, exchange->uri);
+                chosen_exchange, exchange->url);
   }
   if (NULL == exchange)
   {
     /* This is a new exchange */
     exchange = GNUNET_new (struct Exchange);
-    exchange->uri = GNUNET_strdup (chosen_exchange);
+    exchange->url = GNUNET_strdup (chosen_exchange);
     exchange->pending = GNUNET_YES;
     GNUNET_CONTAINER_DLL_insert (exchange_head,
                                  exchange_tail,
@@ -745,11 +843,12 @@ TMH_EXCHANGES_find_exchange (const char *chosen_exchange,
   GNUNET_CONTAINER_DLL_insert (exchange->fo_head,
                                exchange->fo_tail,
                                fo);
-
+  now = GNUNET_TIME_absolute_get ();
+  (void) GNUNET_TIME_round_abs (&now);
   if ( (GNUNET_YES != exchange->pending) &&
        ( (NULL == fo->wire_method) ||
          (NULL != get_wire_fees (exchange,
-                                 GNUNET_TIME_absolute_get (),
+                                 now,
                                  fo->wire_method)) ) )
   {
     /* We are not currently waiting for a reply, immediately
@@ -766,6 +865,7 @@ TMH_EXCHANGES_find_exchange (const char *chosen_exchange,
   {
     GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
                 "Do not have current key data. Will request /keys now\n");
+    GNUNET_assert (NULL == exchange->retry_task);
     exchange->retry_task = GNUNET_SCHEDULER_add_now (&retry_exchange,
                                                      exchange);
   }
@@ -818,7 +918,7 @@ accept_exchanges (void *cls,
                   const char *section)
 {
   const struct GNUNET_CONFIGURATION_Handle *cfg = cls;
-  char *uri;
+  char *url;
   char *mks;
   struct Exchange *exchange;
 
@@ -829,16 +929,16 @@ accept_exchanges (void *cls,
   if (GNUNET_OK !=
       GNUNET_CONFIGURATION_get_value_string (cfg,
                                              section,
-                                             "URI",
-                                             &uri))
+                                             "URL",
+                                             &url))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                section,
-                               "URI");
+                               "URL");
     return;
   }
   exchange = GNUNET_new (struct Exchange);
-  exchange->uri = uri;
+  exchange->url = url;
   if (GNUNET_OK ==
       GNUNET_CONFIGURATION_get_value_string (cfg,
                                              section,
@@ -872,6 +972,7 @@ accept_exchanges (void *cls,
                                exchange_tail,
                                exchange);
   exchange->pending = GNUNET_YES;
+  GNUNET_assert (NULL == exchange->retry_task);
   exchange->retry_task = GNUNET_SCHEDULER_add_now (&retry_exchange,
                                                    exchange);
 }
@@ -887,9 +988,6 @@ accept_exchanges (void *cls,
 int
 TMH_EXCHANGES_init (const struct GNUNET_CONFIGURATION_Handle *cfg)
 {
-  struct Exchange *exchange;
-  json_t *j_exchange;
-
   merchant_curl_ctx
     = GNUNET_CURL_init (&GNUNET_CURL_gnunet_scheduler_reschedule,
                         &merchant_curl_rc);
@@ -905,12 +1003,16 @@ TMH_EXCHANGES_init (const struct GNUNET_CONFIGURATION_Handle *cfg)
                                          (void *) cfg);
   /* build JSON with list of trusted exchanges (will be included in contracts) */
   trusted_exchanges = json_array ();
-  for (exchange = exchange_head; NULL != exchange; exchange = exchange->next)
+  for (struct Exchange *exchange = exchange_head;
+       NULL != exchange;
+       exchange = exchange->next)
   {
+    json_t *j_exchange;
+
     if (GNUNET_YES != exchange->trusted)
       continue;
     j_exchange = json_pack ("{s:s, s:o}",
-                            "url", exchange->uri,
+                            "url", exchange->url,
                             "master_pub", GNUNET_JSON_from_data_auto (&exchange->master_pub));
     GNUNET_assert (0 ==
                    json_array_append_new (trusted_exchanges,
@@ -970,7 +1072,7 @@ TMH_EXCHANGES_done ()
       GNUNET_SCHEDULER_cancel (exchange->retry_task);
       exchange->retry_task = NULL;
     }
-    GNUNET_free (exchange->uri);
+    GNUNET_free (exchange->url);
     GNUNET_free (exchange);
   }
   GNUNET_CURL_fini (merchant_curl_ctx);
