@@ -64,7 +64,7 @@ struct Buffer
 
 
 /**
- * Initialize a buffer.
+ * Initialize a buffer and copy first chunk of data in it.
  *
  * @param buf the buffer to initialize
  * @param data the initial data
@@ -133,6 +133,7 @@ buffer_append (struct Buffer *buf,
   {
     char *new_buf;
     size_t new_size = buf->alloc ? buf->alloc : 1;
+    
     while (new_size < buf->fill + data_size)
       new_size *= 2;
     if (new_size > max_size)
@@ -143,7 +144,9 @@ buffer_append (struct Buffer *buf,
     buf->data = new_buf;
     buf->alloc = new_size;
   }
-  memcpy (buf->data + buf->fill, data, data_size);
+  memcpy (buf->data + buf->fill,
+	  data,
+	  data_size);
   buf->fill += data_size;
   return GNUNET_OK;
 }
@@ -186,7 +189,7 @@ TMH_PARSE_post_cleanup_callback (void *con_cls)
  *               may be parsed in the future (call again);
  *               `*json` will be NULL if we need to be called again,
  *                and non-NULL if we are done.
- *    #GNUNET_NO is request incomplete or invalid
+ *    #GNUNET_NO if request is incomplete or invalid
  *               (error message was generated)
  *    #GNUNET_SYSERR on internal error
  *               (we could not even queue an error message,
@@ -200,7 +203,10 @@ TMH_PARSE_post_json (struct MHD_Connection *connection,
                      json_t **json)
 {
   struct Buffer *r = *con_cls;
-
+  
+  TALER_LOG_DEBUG ("Will parse: %.*s\n",
+                   (int) *upload_data_size,
+                   upload_data);
   *json = NULL;
   if (NULL == *con_cls)
   {
@@ -216,21 +222,27 @@ TMH_PARSE_post_json (struct MHD_Connection *connection,
       *con_cls = NULL;
       buffer_deinit (r);
       GNUNET_free (r);
-      return (MHD_NO ==
-              TMH_RESPONSE_reply_internal_error (connection,
-						 TALER_EC_PARSER_OUT_OF_MEMORY,
-                                                 "out of memory"))
-        ? GNUNET_SYSERR : GNUNET_NO;
+      /* return GNUNET_SYSERR if this isn't even
+       * able to generate proper error response.  */
+      return (MHD_NO == TMH_RESPONSE_reply_internal_error
+        (connection,
+         TALER_EC_PARSER_OUT_OF_MEMORY,
+         "out of memory")) ? GNUNET_SYSERR : GNUNET_NO;
     }
     /* everything OK, wait for more POST data */
     *upload_data_size = 0;
     *con_cls = r;
     return GNUNET_YES;
   }
+
+  /* When zero, upload is over.  */
   if (0 != *upload_data_size)
   {
-    /* We are seeing an old request with more data available. */
+    TALER_LOG_INFO ("Parser asking for more data"
+                    ", current data size is %lu\n",
+                    *upload_data_size);
 
+    /* We are seeing an old request with more data available. */
     if (GNUNET_OK !=
         buffer_append (r,
                        upload_data,
@@ -241,17 +253,19 @@ TMH_PARSE_post_json (struct MHD_Connection *connection,
       *con_cls = NULL;
       buffer_deinit (r);
       GNUNET_free (r);
-      return (MHD_NO ==
-              TMH_RESPONSE_reply_request_too_large (connection))
-        ? GNUNET_SYSERR : GNUNET_NO;
+      return (MHD_NO == TMH_RESPONSE_reply_request_too_large
+        (connection)) ? GNUNET_SYSERR : GNUNET_NO;
     }
+
     /* everything OK, wait for more POST data */
     *upload_data_size = 0;
     return GNUNET_YES;
   }
 
+  TALER_LOG_DEBUG ("About to parse: %.*s\n",
+                   (int) r->fill,
+                   r->data);
   /* We have seen the whole request. */
-
   *json = json_loadb (r->data,
                       r->fill,
                       0,
@@ -260,9 +274,8 @@ TMH_PARSE_post_json (struct MHD_Connection *connection,
   {
     GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                 "Failed to parse JSON request body\n");
-    return (MHD_YES ==
-            TMH_RESPONSE_reply_invalid_json (connection))
-      ? GNUNET_NO : GNUNET_SYSERR;
+    return (MHD_YES == TMH_RESPONSE_reply_invalid_json
+      (connection)) ? GNUNET_NO : GNUNET_SYSERR;
   }
   buffer_deinit (r);
   GNUNET_free (r);
@@ -303,6 +316,9 @@ TMH_PARSE_json_data (struct MHD_Connection *connection,
   {
     if (NULL == error_json_name)
       error_json_name = "<no field>";
+    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+                "Parsing failed due to field '%s'\n",
+                error_json_name);
     ret = (MHD_YES ==
            TMH_RESPONSE_reply_json_pack (connection,
                                          MHD_HTTP_BAD_REQUEST,
