@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015, 2016 GNUnet e.V.
+  Copyright (C) 2014-2017 GNUnet e.V.
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -44,7 +44,11 @@ TMH_RESPONSE_make_json (const json_t *json)
 
   json_str = json_dumps (json,
 			 JSON_INDENT(2));
-  GNUNET_assert (NULL != json_str);
+  if (NULL == json_str)
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
   resp = MHD_create_response_from_buffer (strlen (json_str),
 					  json_str,
                                           MHD_RESPMEM_MUST_FREE);
@@ -177,11 +181,10 @@ TMH_RESPONSE_reply_json_pack (struct MHD_Connection *connection,
  * @return a MHD response object
  */
 struct MHD_Response *
-TMH_RESPONSE_make_internal_error (enum TALER_ErrorCode ec,
-				  const char *hint)
+TMH_RESPONSE_make_error (enum TALER_ErrorCode ec,
+			 const char *hint)
 {
-  return TMH_RESPONSE_make_json_pack ("{s:s, s:I, s:s}",
-                                      "error", "internal error",
+  return TMH_RESPONSE_make_json_pack ("{s:I, s:s}",
 				      "code", (json_int_t) ec,
                                       "hint", hint);
 }
@@ -202,8 +205,7 @@ TMH_RESPONSE_reply_internal_error (struct MHD_Connection *connection,
 {
   return TMH_RESPONSE_reply_json_pack (connection,
                                        MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                       "{s:s, s:I, s:s}",
-                                       "error", "internal error",
+                                       "{s:I, s:s}",
 				       "code", (json_int_t) ec,
                                        "hint", hint);
 }
@@ -234,6 +236,30 @@ TMH_RESPONSE_reply_request_too_large (struct MHD_Connection *connection)
 }
 
 
+/**
+ * Send a response indicating that we did not find the @a object
+ * needed for the reply.
+ *
+ * @param connection the MHD connection to use
+ * @param response_code response code to use
+ * @param ec error code to return
+ * @param msg human-readable diagnostic
+ * @return a MHD result code
+ */
+int
+TMH_RESPONSE_reply_rc (struct MHD_Connection *connection,
+                       unsigned int response_code,
+                       enum TALER_ErrorCode ec,
+                       const char *msg)
+{
+  return TMH_RESPONSE_reply_json_pack (connection,
+                                       response_code,
+                                       "{s:I, s:s}",
+				       "code", (json_int_t) ec,
+                                       "error", msg);
+}
+
+
 /**
  * Send a response indicating that the JSON was malformed.
  *
@@ -268,7 +294,7 @@ TMH_RESPONSE_reply_not_found (struct MHD_Connection *connection,
   return TMH_RESPONSE_reply_json_pack (connection,
                                        MHD_HTTP_NOT_FOUND,
                                        "{s:I, s:s}",
-				       "code", (json_int_t) ec,       
+				       "code", (json_int_t) ec,
                                        "error", object);
 }
 
@@ -326,31 +352,12 @@ TMH_RESPONSE_reply_external_error (struct MHD_Connection *connection,
 {
   return TMH_RESPONSE_reply_json_pack (connection,
                                        MHD_HTTP_BAD_REQUEST,
-                                       "{s:s, s:I, s:s}",
-                                       "error", "client error",
+                                       "{s:I, s:s}",
 				       "code", (json_int_t) ec,
                                        "hint", hint);
 }
 
 
-/**
- * Create a response indicating an external error.
- *
- * @param ec error code to return
- * @param hint hint about the internal error's nature
- * @return a MHD response object
- */
-struct MHD_Response *
-TMH_RESPONSE_make_external_error (enum TALER_ErrorCode ec,
-				  const char *hint)
-{
-  return TMH_RESPONSE_make_json_pack ("{s:s, s:I, s:s}",
-                                      "error", "client error",
-				      "code", (json_int_t) ec,
-                                      "hint", hint);
-}
-
-
 /**
  * Send a response indicating a missing argument.
  *
@@ -366,7 +373,7 @@ TMH_RESPONSE_reply_arg_missing (struct MHD_Connection *connection,
 {
   return TMH_RESPONSE_reply_json_pack (connection,
                                        MHD_HTTP_BAD_REQUEST,
-                                       "{ s:s, s:I, s:s}",
+                                       "{s:s, s:I, s:s}",
                                        "error", "missing parameter",
 				       "code", (json_int_t) ec,
                                        "parameter", param_name);
@@ -395,52 +402,4 @@ TMH_RESPONSE_reply_arg_invalid (struct MHD_Connection *connection,
 }
 
 
-/**
- * Generate /track/transaction response.
- *
- * @param num_transfers how many wire transfers make up the transaction
- * @param transfers data on each wire transfer
- * @param exchange_uri URI of the exchange that made the transfer
- * @return MHD response object
- */
-struct MHD_Response *
-TMH_RESPONSE_make_track_transaction_ok (unsigned int num_transfers,
-                                        const struct TALER_MERCHANT_TransactionWireTransfer *transfers,
-                                        const char *exchange_uri)
-{
-  struct MHD_Response *ret;
-  unsigned int i;
-  json_t *j_transfers;
-  struct TALER_Amount sum;
-
-  j_transfers = json_array ();
-  for (i=0;i<num_transfers;i++)
-  {
-    const struct TALER_MERCHANT_TransactionWireTransfer *transfer = &transfers[i];
-    unsigned int j;
-
-    sum = transfer->coins[0].amount_with_fee;
-    for (j=1;j<transfer->num_coins;j++)
-    {
-      const struct TALER_MERCHANT_CoinWireTransfer *coin = &transfer->coins[j];
-
-      GNUNET_assert (GNUNET_SYSERR != TALER_amount_add (&sum,
-                                                        &sum,
-                                                        &coin->amount_with_fee));
-    }
-
-    GNUNET_assert (0 ==
-                   json_array_append_new (j_transfers,
-                                          json_pack ("{s:s, s:o, s:o, s:o}",
-                                                     "exchange", exchange_uri,
-                                                     "wtid", GNUNET_JSON_from_data_auto (&transfer->wtid),
-                                                     "execution_time", GNUNET_JSON_from_time_abs (transfer->execution_time),
-                                                     "amount", TALER_JSON_from_amount (&sum))));
-  }
-  ret = TMH_RESPONSE_make_json (j_transfers);
-  json_decref (j_transfers);
-  return ret;
-}
-
-
 /* end of taler-exchange-httpd_responses.c */
