@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015, 2016 GNUnet e.V.
+  Copyright (C) 2014-2017 GNUnet e.V.
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -69,20 +69,6 @@ TMH_RESPONSE_make_json_pack (const char *fmt,
                              ...);
 
 
-/**
- * Generate /track/transaction response.
- *
- * @param num_transfers how many wire transfers make up the transaction
- * @param transfers data on each wire transfer
- * @param exchange_uri URI of the exchange that made the transfer
- * @return MHD response object
- */
-struct MHD_Response *
-TMH_RESPONSE_make_track_transaction_ok (unsigned int num_transfers,
-                                        const struct TALER_MERCHANT_TransactionWireTransfer *transfers,
-                                        const char *exchange_uri);
-
-
 /**
  * Function to call to handle the request by building a JSON
  * reply from a format string and varargs.
@@ -110,6 +96,23 @@ int
 TMH_RESPONSE_reply_invalid_json (struct MHD_Connection *connection);
 
 
+/**
+ * Send a response indicating that we did not find the @a object
+ * needed for the reply.
+ *
+ * @param connection the MHD connection to use
+ * @param response_code response code to use
+ * @param ec error code to return
+ * @param msg human-readable diagnostic
+ * @return a MHD result code
+ */
+int
+TMH_RESPONSE_reply_rc (struct MHD_Connection *connection,
+                       unsigned int response_code,
+                       enum TALER_ErrorCode ec,
+                       const char *msg);
+
+
 /**
  * Send a response indicating that we did not find the @a object
  * needed for the reply.
@@ -139,20 +142,6 @@ TMH_RESPONSE_reply_bad_request (struct MHD_Connection *connection,
                                 const char *issue);
 
 
-/**
- * Send a response indicating an internal error.
- *
- * @param connection the MHD connection to use
- * @param ec error code to return
- * @param hint hint about the internal error's nature
- * @return a MHD result code
- */
-int
-TMH_RESPONSE_reply_internal_error (struct MHD_Connection *connection,
-				   enum TALER_ErrorCode ec,
-                                   const char *hint);
-
-
 /**
  * Create a response indicating an internal error.
  *
@@ -163,7 +152,7 @@ TMH_RESPONSE_reply_internal_error (struct MHD_Connection *connection,
 struct MHD_Response *
 TMH_RESPONSE_make_internal_error (enum TALER_ErrorCode ec,
 				  const char *hint);
-  
+
 
 /**
  * Send a response indicating an external error.
@@ -180,15 +169,29 @@ TMH_RESPONSE_reply_external_error (struct MHD_Connection *connection,
 
 
 /**
- * Create a response indicating an external error.
+ * Send a response indicating an internal error.
+ *
+ * @param connection the MHD connection to use
+ * @param ec error code to return
+ * @param hint hint about the internal error's nature
+ * @return a MHD result code
+ */
+int
+TMH_RESPONSE_reply_internal_error (struct MHD_Connection *connection,
+				   enum TALER_ErrorCode ec,
+                                   const char *hint);
+
+
+/**
+ * Create a response indicating an error.
  *
  * @param ec error code to return
  * @param hint hint about the internal error's nature
  * @return a MHD response object
  */
 struct MHD_Response *
-TMH_RESPONSE_make_external_error (enum TALER_ErrorCode ec,
-				  const char *hint);
+TMH_RESPONSE_make_error (enum TALER_ErrorCode ec,
+			 const char *hint);
 
 
 /**
