@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2014, 2015, 2016 INRIA
+  (C) 2014-2017 INRIA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -32,6 +32,41 @@
 #include "taler-merchant-httpd_track-transaction.h"
 
 
+/**
+ * Information about a wire transfer for a /track/transaction response.
+ */
+struct TransactionWireTransfer
+{
+
+  /**
+   * Wire transfer identifier this struct is about.
+   */
+  struct TALER_WireTransferIdentifierRawP wtid;
+
+  /**
+   * When was this wire transfer executed?
+   */
+  struct GNUNET_TIME_Absolute execution_time;
+
+  /**
+   * Number of coins of the selected transaction that
+   * is covered by this wire transfer.
+   */
+  unsigned int num_coins;
+
+  /**
+   * Information about the coins of the selected transaction
+   * that are part of the wire transfer.
+   */
+  struct TALER_MERCHANT_CoinWireTransfer *coins;
+
+  /**
+   * URL of the exchange that executed the wire transfer.
+   */
+  char *exchange_url;
+};
+
+
 /**
  * Map containing all the known merchant instances
  */
@@ -42,6 +77,56 @@ extern struct GNUNET_CONTAINER_MultiHashMap *by_id_map;
  */
 #define TRACK_TIMEOUT (GNUNET_TIME_relative_multiply (GNUNET_TIME_UNIT_SECONDS, 30))
 
+/**
+ * How often do we retry the simple INSERT database transaction?
+ */
+#define MAX_RETRIES 3
+
+
+/**
+ * Generate /track/transaction response.
+ *
+ * @param num_transfers how many wire transfers make up the transaction
+ * @param transfers data on each wire transfer
+ * @return MHD response object
+ */
+static struct MHD_Response *
+make_track_transaction_ok (unsigned int num_transfers,
+			   const struct TransactionWireTransfer *transfers)
+{
+  struct MHD_Response *ret;
+  json_t *j_transfers;
+  struct TALER_Amount sum;
+
+  j_transfers = json_array ();
+  for (unsigned int i=0;i<num_transfers;i++)
+  {
+    const struct TransactionWireTransfer *transfer = &transfers[i];
+
+    sum = transfer->coins[0].amount_with_fee;
+    for (unsigned int j=1;j<transfer->num_coins;j++)
+    {
+      const struct TALER_MERCHANT_CoinWireTransfer *coin = &transfer->coins[j];
+
+      GNUNET_assert (GNUNET_SYSERR !=
+		     TALER_amount_add (&sum,
+				       &sum,
+				       &coin->amount_with_fee));
+    }
+
+    GNUNET_assert (0 ==
+                   json_array_append_new (j_transfers,
+                                          json_pack ("{s:s, s:o, s:o, s:o}",
+                                                     "exchange", transfer->exchange_url,
+                                                     "wtid", GNUNET_JSON_from_data_auto (&transfer->wtid),
+                                                     "execution_time", GNUNET_JSON_from_time_abs (transfer->execution_time),
+                                                     "amount", TALER_JSON_from_amount (&sum))));
+  }
+  ret = TMH_RESPONSE_make_json (j_transfers);
+  json_decref (j_transfers);
+  return ret;
+}
+
 
 /**
  * Context for a /track/transaction operation.
@@ -69,7 +154,7 @@ struct TrackCoinContext
   struct TrackCoinContext *prev;
 
   /**
-   * Our Context for a /track/transaction operation.
+   * Our context for a /track/transaction operation.
    */
   struct TrackTransactionContext *tctx;
 
@@ -78,6 +163,11 @@ struct TrackCoinContext
    */
   struct TALER_CoinSpendPublicKeyP coin_pub;
 
+  /**
+   * Exchange that was used for the transaction.
+   */
+  char *exchange_url;
+
   /**
    * Handle for the request to resolve the WTID for this coin.
    */
@@ -137,11 +227,6 @@ struct TrackTransactionContext
    */
   struct TrackCoinContext *tcc_tail;
 
-  /**
-   * Exchange that was used for the transaction.
-   */
-  char *exchange_uri;
-
   /**
    * Task run on timeout.
    */
@@ -159,6 +244,11 @@ struct TrackTransactionContext
    */
   struct TALER_EXCHANGE_Handle *eh;
 
+  /**
+   * URL of the exchange we currently have in @e eh.
+   */
+  const char *current_exchange;
+
   /**
    * Handle we use to resolve transactions for a given WTID.
    */
@@ -219,6 +309,12 @@ struct TrackTransactionContext
    */
   struct MerchantInstance *mi;
 
+  /**
+   * Set to negative values in #coin_cb() if we encounter
+   * a database problem.
+   */
+  enum GNUNET_DB_QueryStatus qs;
+
 };
 
 
@@ -242,6 +338,11 @@ free_tctx (struct TrackTransactionContext *tctx)
       TALER_EXCHANGE_track_transaction_cancel (tcc->dwh);
       tcc->dwh = NULL;
     }
+    if (NULL != tcc->exchange_url)
+    {
+      GNUNET_free (tcc->exchange_url);
+      tcc->exchange_url = NULL;
+    }
     GNUNET_free (tcc);
   }
   if (NULL != tctx->wdh)
@@ -259,11 +360,6 @@ free_tctx (struct TrackTransactionContext *tctx)
     GNUNET_SCHEDULER_cancel (tctx->timeout_task);
     tctx->timeout_task = NULL;
   }
-  if (NULL != tctx->exchange_uri)
-  {
-    GNUNET_free (tctx->exchange_uri);
-    tctx->exchange_uri = NULL;
-  }
   GNUNET_free (tctx);
 }
 
@@ -358,8 +454,7 @@ wire_deposits_cb (void *cls,
                   const struct TALER_TrackTransferDetails *details)
 {
   struct TrackTransactionContext *tctx = cls;
-  struct TrackCoinContext *tcc;
-  unsigned int i;
+  enum GNUNET_DB_QueryStatus qs;
 
   tctx->wdh = NULL;
   if (MHD_HTTP_OK != http_status)
@@ -375,26 +470,37 @@ wire_deposits_cb (void *cls,
                                     "details", json));
     return;
   }
-  if (GNUNET_OK !=
-      db->store_transfer_to_proof (db->cls,
-                                   tctx->exchange_uri,
-                                   &tctx->current_wtid,
-                                   tctx->current_execution_time,
-                                   exchange_pub,
-                                   json))
+  for (unsigned int i=0;i<MAX_RETRIES;i++)
+  {
+    qs = db->store_transfer_to_proof (db->cls,
+				      tctx->current_exchange,
+				      &tctx->current_wtid,
+				      tctx->current_execution_time,
+				      exchange_pub,
+				      json);
+    if (GNUNET_DB_STATUS_SOFT_ERROR != qs)
+      break;
+  }
+  if (0 > qs)
   {
     /* Not good, but not fatal either, log error and continue */
+    /* Special report if retries insufficient */
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+    /* Always report on hard error as well to enable diagnostics */
+    GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                 "Failed to store transfer-to-proof mapping in DB\n");
   }
-  for (tcc=tctx->tcc_head;NULL != tcc;tcc=tcc->next)
+  for (struct TrackCoinContext *tcc=tctx->tcc_head;
+       NULL != tcc;
+       tcc=tcc->next)
   {
     if (GNUNET_YES == tcc->have_wtid)
       continue;
-    for (i=0;i<details_length;i++)
+    for (unsigned int d=0;d<details_length;d++)
     {
 
-      if (0 == memcmp (&details[i].coin_pub,
+      if (0 == memcmp (&details[d].coin_pub,
                        &tcc->coin_pub,
                        sizeof (struct TALER_CoinSpendPublicKeyP)))
       {
@@ -403,13 +509,22 @@ wire_deposits_cb (void *cls,
         tcc->have_wtid = GNUNET_YES;
       }
 
-      if (GNUNET_OK !=
-          db->store_coin_to_transfer (db->cls,
-                                      &details[i].h_contract_terms,
-                                      &details[i].coin_pub,
-                                      &tctx->current_wtid))
+      for (unsigned int i=0;i<MAX_RETRIES;i++)
       {
-        /* Not good, but not fatal either, log error and continue */
+	qs = db->store_coin_to_transfer (db->cls,
+					 &details[d].h_contract_terms,
+					 &details[d].coin_pub,
+					 &tctx->current_wtid);
+	if (GNUNET_DB_STATUS_SOFT_ERROR != qs)
+	  break;
+      }
+      if (0 > qs)
+      {
+	/* Not good, but not fatal either, log error and continue */
+	/* Special report if retries insufficient */
+	GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+	/* Always report on hard error as well to enable diagnostics */
+	GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
         GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                     "Failed to store coin-to-transfer mapping in DB\n");
       }
@@ -456,11 +571,11 @@ proof_cb (void *cls,
 
 
 /**
- * Function called with detailed wire transfer data.
- * We were trying to find out in which wire transfer one of the
- * coins was involved in. Now we know. What we do now is first
- * obtain the inverse: all other coins of that wire transfer,
- * which is what we prefer to store.
+ * This function takes the wtid from the coin being tracked
+ * and _track_ it against the exchange.  This way, we know
+ * all the other coins which were aggregated together with
+ * this one.  This way we save further HTTP requests to track
+ * the other coins.
  *
  * @param cls closure with a `struct TrackCoinContext`
  * @param http_status HTTP status code we got, 0 on exchange protocol violation
@@ -487,6 +602,7 @@ wtid_cb (void *cls,
   struct TrackCoinContext *tcc = cls;
   struct TrackTransactionContext *tctx = tcc->tctx;
   struct ProofCheckContext pcc;
+  enum GNUNET_DB_QueryStatus qs;
 
   tcc->dwh = NULL;
   if (MHD_HTTP_OK != http_status)
@@ -508,17 +624,35 @@ wtid_cb (void *cls,
   }
   tctx->current_wtid = *wtid;
   tctx->current_execution_time = execution_time;
+
   pcc.p_ret = NULL;
-  /* WARNING: if two transactions got aggregated under the same
-     WTID, then this branch is always taken (when attempting to
-     track the second transaction). */
-  if (GNUNET_OK ==
-      db->find_proof_by_wtid (db->cls,
-                              tctx->exchange_uri,
-                              wtid,
-                              &proof_cb,
-                              &pcc))
+  /* attempt to find this wtid's track from our database,
+     Will make pcc.p_ret point to a "proof", if one exists. */
+  qs = db->find_proof_by_wtid (db->cls,
+			       tctx->current_exchange,
+			       wtid,
+			       &proof_cb,
+			       &pcc);
+  if (0 > qs)
   {
+    /* Simple select queries should not cause serialization issues */
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+    /* Always report on hard error as well to enable diagnostics */
+    GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
+    resume_track_transaction_with_response
+      (tcc->tctx,
+       MHD_HTTP_INTERNAL_SERVER_ERROR,
+       TMH_RESPONSE_make_error (TALER_EC_TRACK_TRANSACTION_DB_FETCH_FAILED,
+				"Fail to query database about proofs"));
+    return;
+  }
+
+  if (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs)
+  {
+    /* How come this wtid was already stored into the
+       database and _not all_ of its coins were already
+       tracked? Inconsistent state (! At least regarding
+       what the exchange tells us) */
     GNUNET_break_op (0);
     resume_track_transaction_with_response
       (tcc->tctx,
@@ -531,6 +665,7 @@ wtid_cb (void *cls,
                                     "coin_pub", GNUNET_JSON_from_data_auto (&tcc->coin_pub)));
     return;
   }
+
   tctx->wdh = TALER_EXCHANGE_track_transfer (tctx->eh,
                                              wtid,
                                              &wire_deposits_cb,
@@ -539,45 +674,27 @@ wtid_cb (void *cls,
 
 
 /**
- * This function is called to trace the wire transfers for
- * all of the coins of the transaction of the @a tctx.  Once
- * we have traced all coins, we build the response.
+ * We have obtained all WTIDs, now prepare the response
  *
- * @param tctx track context with established connection to exchange
+ * @param tctx handle for the operation
  */
 static void
-trace_coins (struct TrackTransactionContext *tctx)
+generate_response (struct TrackTransactionContext *tctx)
 {
   struct TrackCoinContext *tcc;
   unsigned int num_wtid;
 
-  GNUNET_assert (NULL != tctx->eh);
-  for (tcc = tctx->tcc_head; NULL != tcc; tcc = tcc->next)
-    if (GNUNET_YES != tcc->have_wtid)
-      break;
-  if (NULL != tcc)
-  {
-    /* we are not done requesting WTIDs, do the next one */
-    tcc->dwh = TALER_EXCHANGE_track_transaction (tctx->eh,
-                                                 &tctx->mi->privkey,
-                                                 &tctx->h_wire,
-                                                 &tctx->h_contract_terms,
-                                                 &tcc->coin_pub,
-                                                 &wtid_cb,
-                                                 tcc);
-    return;
-  }
-  /* We have obtained all WTIDs, now prepare the response */
   num_wtid = 0;
   /* count how many disjoint wire transfer identifiers there are;
      note that there should only usually be one, so while this
      is worst-case O(n^2), in pracitce this is O(n) */
   for (tcc = tctx->tcc_head; NULL != tcc; tcc = tcc->next)
   {
-    struct TrackCoinContext *tcc2;
     int found = GNUNET_NO;
 
-    for (tcc2 = tctx->tcc_head; tcc2 != tcc; tcc2 = tcc2->next)
+    for (struct TrackCoinContext *tcc2 = tctx->tcc_head;
+	 tcc2 != tcc;
+	 tcc2 = tcc2->next)
     {
       if (0 == memcmp (&tcc->wtid,
                        &tcc2->wtid,
@@ -595,16 +712,17 @@ trace_coins (struct TrackTransactionContext *tctx)
     /* on-stack allocation is fine, as the number of coins and the
        number of wire-transfers per-transaction is expected to be tiny. */
     struct MHD_Response *resp;
-    struct TALER_MERCHANT_TransactionWireTransfer wts[num_wtid];
+    struct TransactionWireTransfer wts[num_wtid];
     unsigned int wtid_off;
 
     wtid_off = 0;
     for (tcc = tctx->tcc_head; NULL != tcc; tcc = tcc->next)
     {
-      struct TrackCoinContext *tcc2;
       int found = GNUNET_NO;
 
-      for (tcc2 = tctx->tcc_head; tcc2 != tcc; tcc2 = tcc2->next)
+      for (struct TrackCoinContext *tcc2 = tctx->tcc_head;
+	   tcc2 != tcc;
+	   tcc2 = tcc2->next)
       {
         if (0 == memcmp (&tcc->wtid,
                          &tcc2->wtid,
@@ -617,14 +735,17 @@ trace_coins (struct TrackTransactionContext *tctx)
       if (GNUNET_NO == found)
       {
         unsigned int num_coins;
-        struct TALER_MERCHANT_TransactionWireTransfer *wt;
+        struct TransactionWireTransfer *wt;
 
         wt = &wts[wtid_off++];
         wt->wtid = tcc->wtid;
+	wt->exchange_url = tcc->exchange_url;
         wt->execution_time = tcc->execution_time;
         /* count number of coins with this wtid */
         num_coins = 0;
-        for (tcc2 = tctx->tcc_head; NULL != tcc2; tcc2 = tcc2->next)
+        for (struct TrackCoinContext *tcc2 = tctx->tcc_head;
+	     NULL != tcc2;
+	     tcc2 = tcc2->next)
         {
           if (0 == memcmp (&wt->wtid,
                            &tcc2->wtid,
@@ -636,7 +757,9 @@ trace_coins (struct TrackTransactionContext *tctx)
         wt->coins = GNUNET_new_array (num_coins,
                                       struct TALER_MERCHANT_CoinWireTransfer);
         num_coins = 0;
-        for (tcc2 = tctx->tcc_head; NULL != tcc2; tcc2 = tcc2->next)
+        for (struct TrackCoinContext *tcc2 = tctx->tcc_head;
+	     NULL != tcc2;
+	     tcc2 = tcc2->next)
         {
           if (0 == memcmp (&wt->wtid,
                            &tcc2->wtid,
@@ -653,9 +776,8 @@ trace_coins (struct TrackTransactionContext *tctx)
     } /* for all tcc */
     GNUNET_assert (wtid_off == num_wtid);
 
-    resp = TMH_RESPONSE_make_track_transaction_ok (num_wtid,
-                                                   wts,
-                                                   tctx->exchange_uri);
+    resp = make_track_transaction_ok (num_wtid,
+				      wts);
     for (wtid_off=0;wtid_off < num_wtid; wtid_off++)
       GNUNET_free (wts[wtid_off].coins);
     resume_track_transaction_with_response (tctx,
@@ -665,8 +787,68 @@ trace_coins (struct TrackTransactionContext *tctx)
 }
 
 
+/**
+ * Find the exchange to trace the next coin(s).
+ *
+ * @param tctx operation context
+ */
+static void
+find_exchange (struct TrackTransactionContext *tctx);
+
+
+/**
+ * This function is called to 'trace the wire transfers'
+ * (true?) for all of the coins of the transaction of the @a tctx.
+ * Once we have traced all coins, we build the response.
+ *
+ * @param tctx track context with established connection to exchange
+ */
+static void
+trace_coins (struct TrackTransactionContext *tctx)
+{
+  struct TrackCoinContext *tcc;
+
+  /* Make sure we are connected to the exchange. */
+  GNUNET_assert (NULL != tctx->eh);
+
+  for (tcc = tctx->tcc_head; NULL != tcc; tcc = tcc->next)
+
+    /* How come one does't have wtid? */
+    if (GNUNET_YES != tcc->have_wtid)
+      break;
+
+  if (NULL != tcc)
+  {
+    if (0 != strcmp (tcc->exchange_url,
+		     tctx->current_exchange))
+    {
+      /* exchange changed, find matching one first! */
+      tctx->eh = NULL;
+      tctx->current_exchange = NULL;
+      find_exchange (tctx);
+      return;
+    }
+    /* we are not done requesting WTIDs from the current
+       exchange; do the next one */
+    tcc->dwh = TALER_EXCHANGE_track_transaction (tctx->eh,
+                                                 &tctx->mi->privkey,
+                                                 &tctx->h_wire,
+                                                 &tctx->h_contract_terms,
+                                                 &tcc->coin_pub,
+                                                 &wtid_cb,
+                                                 tcc);
+    return;
+  }
+  tctx->current_exchange = NULL;
+  tctx->eh = NULL;
+  generate_response (tctx);
+}
+
+
 /**
  * Function called with the result of our exchange lookup.
+ * Merely provide the execution context to the routine actually
+ * tracking the coin.
  *
  * @param cls the `struct TrackTransactionContext`
  * @param eh NULL if exchange was not found to be acceptable
@@ -708,46 +890,11 @@ handle_track_transaction_timeout (void *cls)
   }
   resume_track_transaction_with_response (tctx,
                                           MHD_HTTP_SERVICE_UNAVAILABLE,
-                                          TMH_RESPONSE_make_internal_error (TALER_EC_PAY_EXCHANGE_TIMEOUT,
+                                          TMH_RESPONSE_make_error (TALER_EC_PAY_EXCHANGE_TIMEOUT,
 									    "exchange not reachable"));
 }
 
 
-/**
- * Function called with information about a transaction.
- * Its duty is to fill up the "context" for the whole operation.
- *
- * @param cls closure
- * @param transaction_id of the contract
- * @param merchant's public key
- * @param exchange_uri URI of the exchange
- * @param h_contract hash of the contract
- * @param h_wire hash of our wire details
- * @param timestamp time of the confirmation
- * @param refund refund deadline
- * @param total_amount total amount we receive for the contract after fees
- */
-static void
-transaction_cb (void *cls,
-		const struct TALER_MerchantPublicKeyP *merchant_pub,
-                const char *exchange_uri,
-                const struct GNUNET_HashCode *h_contract_terms,
-                const struct GNUNET_HashCode *h_wire,
-                struct GNUNET_TIME_Absolute timestamp,
-                struct GNUNET_TIME_Absolute refund,
-                const struct TALER_Amount *total_amount)
-{
-  struct TrackTransactionContext *tctx = cls;
-
-  tctx->h_contract_terms = *h_contract_terms;
-  tctx->exchange_uri = GNUNET_strdup (exchange_uri);
-  tctx->h_wire = *h_wire;
-  tctx->timestamp = timestamp;
-  tctx->refund_deadline = refund;
-  tctx->total_amount = *total_amount;
-}
-
-
 /**
  * Information about the wire transfer corresponding to
  * a deposit operation.  Note that it is in theory possible
@@ -767,7 +914,7 @@ transaction_cb (void *cls,
  */
 static void
 transfer_cb (void *cls,
-             const struct GNUNET_HashCode *h_contract_terms,
+	     const struct GNUNET_HashCode *h_contract_terms,
              const struct TALER_CoinSpendPublicKeyP *coin_pub,
              const struct TALER_WireTransferIdentifierRawP *wtid,
              struct GNUNET_TIME_Absolute execution_time,
@@ -786,41 +933,88 @@ transfer_cb (void *cls,
 
 
 /**
- * Function called with information about a coin that was deposited.
+ * Responsible to get the current coin wtid and store it into its state.
  *
  * @param cls closure
  * @param transaction_id of the contract
  * @param coin_pub public key of the coin
+ * @param exchange_url URL of exchange that issued @a coin_pub
  * @param amount_with_fee amount the exchange will deposit for this coin
  * @param deposit_fee fee the exchange will charge for this coin
+ * @param refund_fee fee the exchange will charge for refunding this coin
  * @param exchange_proof proof from exchange that coin was accepted
  */
 static void
 coin_cb (void *cls,
          const struct GNUNET_HashCode *h_contract_terms,
          const struct TALER_CoinSpendPublicKeyP *coin_pub,
+	 const char *exchange_url,
          const struct TALER_Amount *amount_with_fee,
          const struct TALER_Amount *deposit_fee,
+         const struct TALER_Amount *refund_fee,
+         const struct TALER_Amount *wire_fee,
          const json_t *exchange_proof)
 {
   struct TrackTransactionContext *tctx = cls;
   struct TrackCoinContext *tcc;
+  enum GNUNET_DB_QueryStatus qs;
 
   tcc = GNUNET_new (struct TrackCoinContext);
   tcc->tctx = tctx;
   tcc->coin_pub = *coin_pub;
+  tcc->exchange_url = GNUNET_strdup (exchange_url);
+
   tcc->amount_with_fee = *amount_with_fee;
   tcc->deposit_fee = *deposit_fee;
   GNUNET_CONTAINER_DLL_insert (tctx->tcc_head,
                                tctx->tcc_tail,
                                tcc);
-  GNUNET_break (GNUNET_SYSERR !=
-                db->find_transfers_by_hash (db->cls,
-                                            h_contract_terms,
-                                            &transfer_cb,
-                                            tcc));
+
+  /* find all those <coin, wtid> pairs associated to
+     this contract term's hash code.  The callback
+     will then set the wtid for the "current coin"
+     context. */
+  qs = db->find_transfers_by_hash (db->cls,
+				   h_contract_terms,
+				   &transfer_cb,
+				   tcc);
+  if (0 > qs)
+  {
+    GNUNET_break (0);
+    tctx->qs = qs;
+  }
 }
 
+
+/**
+ * Find the exchange to trace the next coin(s).
+ *
+ * @param tctx operation context
+ */
+static void
+find_exchange (struct TrackTransactionContext *tctx)
+{
+  struct TrackCoinContext *tcc = tctx->tcc_head;
+
+  while ( (NULL != tcc) &&
+	  (GNUNET_YES == tcc->have_wtid) )
+    tcc = tcc->next;
+  if (NULL != tcc)
+  {
+    tctx->current_exchange = tcc->exchange_url;
+    tctx->fo = TMH_EXCHANGES_find_exchange (tctx->current_exchange,
+					    NULL,
+					    &process_track_transaction_with_exchange,
+					    tctx);
+
+  }
+  else
+  {
+    generate_response (tctx);
+  }
+}
+
+
 /**
  * Handle a "/track/transaction" request.
  *
@@ -842,9 +1036,10 @@ MH_handler_track_transaction (struct TMH_RequestHandler *rh,
   const char *order_id;
   const char *instance;
   int ret;
+  enum GNUNET_DB_QueryStatus qs;
   struct GNUNET_HashCode h_instance;
-  struct GNUNET_HashCode h_contract_terms;
   struct json_t *contract_terms;
+  char *last_session_id;
 
   if (NULL == *connection_cls)
   {
@@ -916,80 +1111,104 @@ MH_handler_track_transaction (struct TMH_RequestHandler *rh,
     return TMH_RESPONSE_reply_not_found (connection,
 					 TALER_EC_TRACK_TRANSACTION_INSTANCE_UNKNOWN,
 					 "unknown instance");
-
-  if (GNUNET_YES !=
-      db->find_contract_terms (db->cls,
-                               &contract_terms,
-                               order_id,
-                               &tctx->mi->pubkey))
+  
+  /* Map order id to contract terms; the objective is to get
+     the contract term's hashcode so as to retrieve all the
+     coins which have been deposited for it. */
+  qs = db->find_contract_terms (db->cls,
+				&contract_terms,
+                                &last_session_id,
+				order_id,
+				&tctx->mi->pubkey);
+  if (0 > qs)
+  {
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+    return TMH_RESPONSE_reply_internal_error (connection,
+					      TALER_EC_TRACK_TRANSACTION_DB_FETCH_TRANSACTION_ERROR,
+                                              "Database error finding contract terms");
+  }
+  if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
     return TMH_RESPONSE_reply_not_found (connection,
 					 TALER_EC_PROPOSAL_LOOKUP_NOT_FOUND,
 					 "Given order_id doesn't map to any proposal");
-  TALER_JSON_hash (contract_terms,
-                   &h_contract_terms);
-  json_decref (contract_terms);
 
-  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
-              "Trying to track h_contract_terms '%s'\n",
-              GNUNET_h2s (&h_contract_terms));
+  GNUNET_free (last_session_id);
 
-  ret = db->find_transaction (db->cls,
-                              &h_contract_terms,
-			      &tctx->mi->pubkey,
-                              &transaction_cb,
-                              tctx);
-  if (GNUNET_NO == ret)
+  if (GNUNET_OK !=
+      TALER_JSON_hash (contract_terms,
+                       &tctx->h_contract_terms))
   {
-    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
-                "h_contract_terms not found\n");
-    return TMH_RESPONSE_reply_not_found (connection,
-					 TALER_EC_TRACK_TRANSACTION_TRANSACTION_UNKNOWN,
-                                         "h_contract_terms is unknown");
-  }
-  if ( (GNUNET_SYSERR == ret) ||
-       (0 != memcmp (&tctx->h_contract_terms,
-                     &h_contract_terms,
-                     sizeof (struct GNUNET_HashCode))) ||
-       (NULL == tctx->exchange_uri) )
-  {
-    GNUNET_break (0);
+    json_decref (contract_terms);
     return TMH_RESPONSE_reply_internal_error (connection,
-					      TALER_EC_TRACK_TRANSACTION_DB_FETCH_TRANSACTION_ERROR,
-                                              "Database error");
+					      TALER_EC_INTERNAL_LOGIC_ERROR,
+                                              "Failed to hash contract terms");
   }
-  ret = db->find_payments (db->cls,
-                           &h_contract_terms,
-                           &tctx->mi->pubkey,
-                           &coin_cb,
-                           tctx);
-  if (GNUNET_SYSERR == ret)
+
   {
-    GNUNET_break (0);
+    struct GNUNET_JSON_Specification spec[] = {
+      GNUNET_JSON_spec_absolute_time ("refund_deadline",
+                                      &tctx->refund_deadline),
+      GNUNET_JSON_spec_absolute_time ("timestamp",
+                                      &tctx->timestamp),
+      TALER_JSON_spec_amount ("amount",
+                              &tctx->total_amount),
+      GNUNET_JSON_spec_fixed_auto ("H_wire",
+                                   &tctx->h_wire),
+      GNUNET_JSON_spec_end()
+    };
+
+    if (GNUNET_YES !=
+        GNUNET_JSON_parse (contract_terms,
+                           spec,
+                           NULL, NULL))
+    {
+      GNUNET_break (0);
+      GNUNET_JSON_parse_free (spec);
+      json_decref (contract_terms);
+      return TMH_RESPONSE_reply_internal_error (connection,
+                                                TALER_EC_INTERNAL_LOGIC_ERROR,
+                                                "Failed to parse contract terms from DB");
+    }
+    json_decref (contract_terms);
+  }
+
+  tctx->qs = GNUNET_DB_STATUS_SUCCESS_NO_RESULTS;
+  db->preflight (db->cls);
+
+  /* Find coins which have been deposited for this contract,
+     and retrieve the wtid for each one. */
+  qs = db->find_payments (db->cls,
+			  &tctx->h_contract_terms,
+			  &tctx->mi->pubkey,
+			  &coin_cb,
+			  tctx);
+  if ( (0 > qs) ||
+       (0 > tctx->qs) )
+  {
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != tctx->qs);
     return TMH_RESPONSE_reply_internal_error (connection,
                                               TALER_EC_TRACK_TRANSACTION_DB_FETCH_PAYMENT_ERROR,
-					      "Database error");
+					      "Database error: failed to find payment data");
   }
-  if (GNUNET_NO == ret)
+  if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
   {
     return TMH_RESPONSE_reply_not_found (connection,
 					 TALER_EC_TRACK_TRANSACTION_DB_NO_DEPOSITS_ERROR,
-                                         "deposits");
+                                         "deposit data not found");
   }
   *connection_cls = tctx;
 
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
               "Suspending /track/transaction handling while working with the exchange\n");
   MHD_suspend_connection (connection);
-  tctx->fo = TMH_EXCHANGES_find_exchange (tctx->exchange_uri,
-                                          NULL,
-                                          &process_track_transaction_with_exchange,
-                                          tctx);
-
   tctx->timeout_task
     = GNUNET_SCHEDULER_add_delayed (TRACK_TIMEOUT,
 				    &handle_track_transaction_timeout,
 				    tctx);
+  find_exchange (tctx);
   return MHD_YES;
 }
 
+
 /* end of taler-merchant-httpd_track-transaction.c */
