@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2014, 2015, 2016 INRIA
+  (C) 2014-2017 INRIA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -37,6 +37,10 @@
  */
 #define TRACK_TIMEOUT (GNUNET_TIME_relative_multiply (GNUNET_TIME_UNIT_SECONDS, 30))
 
+/**
+ * How often do we retry the simple INSERT database transaction?
+ */
+#define MAX_RETRIES 3
 
 /**
  * Context used for handing /track/transfer requests.
@@ -87,9 +91,14 @@ struct TrackTransferContext
   struct GNUNET_SCHEDULER_Task *timeout_task;
 
   /**
-   * URI of the exchange.
+   * URL of the exchange.
    */
-  char *uri;
+  char *url;
+
+  /**
+   * Wire method used for the transfer.
+   */
+  char *wire_method;
 
   /**
    * Pointer to the detail that we are currently
@@ -130,11 +139,13 @@ struct TrackTransferContext
   int check_transfer_result;
 };
 
+
 /**
  * Represents an entry in the table used to sum up
  * individual deposits for each h_contract_terms.
  */
-struct Entry {
+struct Entry
+{
 
   /**
    * Sum accumulator for deposited value.
@@ -172,10 +183,15 @@ free_transfer_track_context (struct TrackTransferContext *rctx)
     TALER_EXCHANGE_track_transfer_cancel (rctx->wdh);
     rctx->wdh = NULL;
   }
-  if (NULL != rctx->uri)
+  if (NULL != rctx->url)
   {
-    GNUNET_free (rctx->uri);
-    rctx->uri = NULL;
+    GNUNET_free (rctx->url);
+    rctx->url = NULL;
+  }
+  if (NULL != rctx->wire_method)
+  {
+    GNUNET_free (rctx->wire_method);
+    rctx->wire_method = NULL;
   }
   GNUNET_free (rctx);
 }
@@ -195,8 +211,8 @@ hashmap_free (void *cls,
               void *value)
 {
   struct TALER_Amount *amount = value;
+
   GNUNET_free (amount);
-  /*NOTE: how to find out when iteration should stop?*/
   return GNUNET_YES;
 }
 
@@ -305,8 +321,9 @@ transform_response (const json_t *result,
     GNUNET_CRYPTO_hash_from_string (key,
                                     &h_key);
 
-    if (NULL != (current_entry = GNUNET_CONTAINER_multihashmap_get (map,
-                                                                    &h_key)))
+    if (NULL != (current_entry =
+		 GNUNET_CONTAINER_multihashmap_get (map,
+						    &h_key)))
     {
       /* The map already knows this h_contract_terms*/
       if ( (GNUNET_SYSERR ==
@@ -409,23 +426,28 @@ track_transfer_cleanup (struct TM_HandlerContext *hc)
 
 
 /**
- * Function called with information about a coin that was transfered.
- * Verify that it matches the information claimed by the exchange.
- * Update the `check_transfer_result` field accordingly.
+ * This function checks that the information about the coin which
+ * was paid back by _this_ wire transfer matches what _we_ (the merchant)
+ * knew about this coin.
  *
  * @param cls closure with our `struct TrackTransferContext *`
  * @param transaction_id of the contract
  * @param coin_pub public key of the coin
+ * @param exchange_url URL of the exchange that issued @a coin_pub
  * @param amount_with_fee amount the exchange will transfer for this coin
  * @param deposit_fee fee the exchange will charge for this coin
+ * @param refund_fee fee the exchange will charge for refunding this coin
  * @param exchange_proof proof from exchange that coin was accepted
  */
 static void
 check_transfer (void *cls,
                 const struct GNUNET_HashCode *h_contract_terms,
                 const struct TALER_CoinSpendPublicKeyP *coin_pub,
+		const char *exchange_url,
                 const struct TALER_Amount *amount_with_fee,
                 const struct TALER_Amount *deposit_fee,
+                const struct TALER_Amount *refund_fee,
+                const struct TALER_Amount *wire_fee,
                 const json_t *exchange_proof)
 {
   struct TrackTransferContext *rctx = cls;
@@ -460,6 +482,90 @@ check_transfer (void *cls,
 }
 
 
+/**
+ * Check that the given @a wire_fee is what the
+ * @a exchange_pub should charge at the @a execution_time.
+ * If the fee is correct (according to our database),
+ * return #GNUNET_OK.  If we do not have the fee structure
+ * in our DB, we just accept it and return #GNUNET_NO;
+ * if we have proof that the fee is bogus, we respond with
+ * the proof to the client and return #GNUNET_SYSERR.
+ *
+ * @param rctx context of the transfer to respond to
+ * @param json response from the exchange
+ * @param execution_time time of the wire transfer
+ * @param wire_fee fee claimed by the exchange
+ * @return #GNUNET_SYSERR if we returned hard proof of
+ *   missbehavior from the exchange to the client
+ */
+static int
+check_wire_fee (struct TrackTransferContext *rctx,
+		const json_t *json,
+		struct GNUNET_TIME_Absolute execution_time,
+		const struct TALER_Amount *wire_fee)
+{
+  const struct TALER_MasterPublicKeyP *master_pub;
+  struct GNUNET_HashCode h_wire_method;
+  struct TALER_Amount expected_fee;
+  struct TALER_Amount closing_fee;
+  struct TALER_MasterSignatureP master_sig;
+  struct GNUNET_TIME_Absolute start_date;
+  struct GNUNET_TIME_Absolute end_date;
+  enum GNUNET_DB_QueryStatus qs;
+  const struct TALER_EXCHANGE_Keys *keys;
+
+  keys = TALER_EXCHANGE_get_keys (rctx->eh);
+  if (NULL == keys)
+  {
+    GNUNET_break (0);
+    return GNUNET_NO;
+  }
+  master_pub = &keys->master_pub;
+  GNUNET_CRYPTO_hash (rctx->wire_method,
+		      strlen (rctx->wire_method) + 1,
+		      &h_wire_method);
+  qs = db->lookup_wire_fee (db->cls,
+			    master_pub,
+			    &h_wire_method,
+			    execution_time,
+			    &expected_fee,
+			    &closing_fee,
+			    &start_date,
+			    &end_date,
+			    &master_sig);
+  if (0 >= qs)
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+		"Failed to find wire fee for `%s' and method `%s' at %s in DB, accepting blindly that the fee is %s\n",
+		TALER_B2S (master_pub),
+		rctx->wire_method,
+		GNUNET_STRINGS_absolute_time_to_string (execution_time),
+		TALER_amount2s (wire_fee));
+    return GNUNET_NO;
+  }
+  if (0 <= TALER_amount_cmp (&expected_fee,
+			     wire_fee))
+    return GNUNET_OK; /* expected_fee >= wire_fee */
+
+  /* Wire fee check failed, export proof to client */
+  resume_track_transfer_with_response
+    (rctx,
+     MHD_HTTP_INTERNAL_SERVER_ERROR,
+     TMH_RESPONSE_make_json_pack ("{s:I, s:o, s:o, s:o, s:o, s:o, s:o, s:o, s:o, s:O}",
+				  "code", (json_int_t) TALER_EC_TRACK_TRANSFER_JSON_BAD_WIRE_FEE,
+				  "wire_fee", TALER_JSON_from_amount (wire_fee),
+				  "execution_time", GNUNET_JSON_from_time_abs (execution_time),
+				  "expected_wire_fee", TALER_JSON_from_amount (&expected_fee),
+				  "expected_closing_fee", TALER_JSON_from_amount (&closing_fee),
+				  "start_date", GNUNET_JSON_from_time_abs (start_date),
+				  "end_date", GNUNET_JSON_from_time_abs (end_date),
+				  "master_sig", GNUNET_JSON_from_data_auto (&master_sig),
+				  "master_pub", GNUNET_JSON_from_data_auto (master_pub),
+				  "json", json));
+  return GNUNET_SYSERR;
+}
+
+
 /**
  * Function called with detailed wire transfer data, including all
  * of the coin transactions that were combined into the wire transfer.
@@ -492,9 +598,8 @@ wire_transfer_cb (void *cls,
                   const struct TALER_TrackTransferDetails *details)
 {
   struct TrackTransferContext *rctx = cls;
-  unsigned int i;
-  int ret;
   json_t *jresponse;
+  enum GNUNET_DB_QueryStatus qs;
 
   rctx->wdh = NULL;
   GNUNET_log (GNUNET_ERROR_TYPE_INFO,
@@ -512,17 +617,23 @@ wire_transfer_cb (void *cls,
                                     "details", json));
     return;
   }
-
-  if (GNUNET_OK !=
-      db->store_transfer_to_proof (db->cls,
-                                   rctx->uri,
-                                   &rctx->wtid,
-                                   execution_time,
-                                   exchange_pub,
-                                   json))
+  for (unsigned int i=0;i<MAX_RETRIES;i++)
   {
-    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                "Failed to persist wire transfer proof in DB\n");
+    qs = db->store_transfer_to_proof (db->cls,
+				      rctx->url,
+				      &rctx->wtid,
+				      execution_time,
+				      exchange_pub,
+				      json);
+    if (GNUNET_DB_STATUS_SOFT_ERROR != qs)
+      break;
+  }
+  if (0 > qs)
+  {
+    /* Special report if retries insufficient */
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+    /* Always report on hard error as well to enable diagnostics */
+    GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
     resume_track_transfer_with_response
       (rctx,
        MHD_HTTP_INTERNAL_SERVER_ERROR,
@@ -532,21 +643,41 @@ wire_transfer_cb (void *cls,
     return;
   }
   rctx->original_response = json;
-  for (i=0;i<details_length;i++)
+
+  if (GNUNET_SYSERR ==
+      check_wire_fee (rctx,
+		      json,
+		      execution_time,
+		      wire_fee))
+    return;
+
+  /* Now we want to double-check that any (Taler coin) deposit
+   * which is accounted into _this_ wire transfer, does exist
+   * into _our_ database.  This is the rationale: if the
+   * exchange paid us for it, we must have received it _beforehands_!
+   *
+   * details_length is how many (Taler coin) deposits have been
+   * aggregated into _this_ wire transfer.
+   */
+  for (unsigned int i=0;i<details_length;i++)
   {
     rctx->current_offset = i;
     rctx->current_detail = &details[i];
+    /* Set the coin as "never seen" before. */
     rctx->check_transfer_result = GNUNET_NO;
-    ret = db->find_payments_by_hash_and_coin (db->cls,
-                                              &details[i].h_contract_terms,
-                                              &rctx->mi->pubkey,
-                                              &details[i].coin_pub,
-                                              &check_transfer,
-                                              rctx);
-    if (GNUNET_SYSERR == ret)
+    qs = db->find_payments_by_hash_and_coin (db->cls,
+					     &details[i].h_contract_terms,
+					     &rctx->mi->pubkey,
+					     &details[i].coin_pub,
+					     &check_transfer,
+					     rctx);
+    if (0 > qs)
     {
-      GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                  "Failed to obtain existing payment data from DB\n");
+      /* single, read-only SQL statements should never cause
+	 serialization problems */
+      GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+      /* Always report on hard error as well to enable diagnostics */
+      GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
       resume_track_transfer_with_response
         (rctx,
          MHD_HTTP_INTERNAL_SERVER_ERROR,
@@ -555,11 +686,11 @@ wire_transfer_cb (void *cls,
                                       "details", "failed to obtain deposit data from local database"));
       return;
     }
-    if (GNUNET_NO == ret)
+    if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
     {
       /* The exchange says we made this deposit, but WE do not
-         recall making it! Well, let's say thanks and accept the
-         money! */
+         recall making it (corrupted / unreliable database?)!
+         Well, let's say thanks and accept the money! */
       GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                   "Failed to find payment data in DB\n");
       rctx->check_transfer_result = GNUNET_OK;
@@ -591,16 +722,23 @@ wire_transfer_cb (void *cls,
       rctx->response = NULL;
       return;
     }
-    /* Response is consistent with the /deposit we made, remember
-       it for future reference */
-    ret = db->store_coin_to_transfer (db->cls,
-                                      &details[i].h_contract_terms,
-                                      &details[i].coin_pub,
-                                      &rctx->wtid);
-    if (GNUNET_OK != ret)
+    /* Response is consistent with the /deposit we made,
+       remember it for future reference */
+    for (unsigned int i=0;i<MAX_RETRIES;i++)
     {
-      GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                  "Failed to persist coin to wire transfer mapping in DB\n");
+      qs = db->store_coin_to_transfer (db->cls,
+				       &details[i].h_contract_terms,
+				       &details[i].coin_pub,
+				       &rctx->wtid);
+      if (GNUNET_DB_STATUS_SOFT_ERROR != qs)
+	break;
+    }
+    if (0 > qs)
+    {
+      /* Special report if retries insufficient */
+      GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+      /* Always report on hard error as well to enable diagnostics */
+      GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
       resume_track_transfer_with_response
         (rctx,
          MHD_HTTP_INTERNAL_SERVER_ERROR,
@@ -615,13 +753,15 @@ wire_transfer_cb (void *cls,
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
               "About to call tracks transformator.\n");
 
-  if (NULL == (jresponse = transform_response (json, rctx)))
+  if (NULL == (jresponse =
+	       transform_response (json,
+				   rctx)))
   {
     resume_track_transfer_with_response
       (rctx,
        MHD_HTTP_INTERNAL_SERVER_ERROR,
-       TMH_RESPONSE_make_internal_error (TALER_EC_TRACK_TRANSFER_JSON_RESPONSE_ERROR,
-                                         "Fail to elaborate the response."));
+       TMH_RESPONSE_make_error (TALER_EC_TRACK_TRANSFER_JSON_RESPONSE_ERROR,
+				"Fail to elaborate the response."));
     return;
   }
 
@@ -688,7 +828,7 @@ handle_track_transfer_timeout (void *cls)
   }
   resume_track_transfer_with_response (rctx,
                                        MHD_HTTP_SERVICE_UNAVAILABLE,
-                                       TMH_RESPONSE_make_internal_error (TALER_EC_TRACK_TRANSFER_EXCHANGE_TIMEOUT,
+                                       TMH_RESPONSE_make_error (TALER_EC_TRACK_TRANSFER_EXCHANGE_TIMEOUT,
 									 "exchange not reachable"));
 }
 
@@ -709,12 +849,14 @@ proof_cb (void *cls,
   struct TrackTransferContext *rctx = cls;
   json_t *transformed_response;
 
-  if (NULL == (transformed_response = transform_response (proof,
-                                                          rctx)))
+  if (NULL == (transformed_response =
+	       transform_response (proof,
+				   rctx)))
   {
     rctx->response_code = MHD_HTTP_INTERNAL_SERVER_ERROR;
-    rctx->response = TMH_RESPONSE_make_internal_error (TALER_EC_TRACK_TRANSFER_JSON_RESPONSE_ERROR,
-                                                       "Fail to elaborate response.");
+    rctx->response
+      = TMH_RESPONSE_make_error (TALER_EC_TRACK_TRANSFER_JSON_RESPONSE_ERROR,
+				 "Fail to elaborate response.");
     return;
   }
 
@@ -745,9 +887,11 @@ MH_handler_track_transfer (struct TMH_RequestHandler *rh,
 {
   struct TrackTransferContext *rctx;
   const char *str;
-  const char *uri;
+  const char *url;
   const char *instance_str;
+  const char *wire_method;
   int ret;
+  enum GNUNET_DB_QueryStatus qs;
 
   if (NULL == *connection_cls)
   {
@@ -793,14 +937,34 @@ MH_handler_track_transfer (struct TMH_RequestHandler *rh,
     return MHD_YES; /* still work in progress */
   }
 
-  uri = MHD_lookup_connection_value (connection,
+  url = MHD_lookup_connection_value (connection,
                                      MHD_GET_ARGUMENT_KIND,
                                      "exchange");
-  if (NULL == uri)
+  if (NULL == url)
     return TMH_RESPONSE_reply_arg_missing (connection,
 					   TALER_EC_PARAMETER_MISSING,
                                            "exchange");
-  rctx->uri = GNUNET_strdup (uri);
+  rctx->url = GNUNET_strdup (url);
+
+  wire_method = MHD_lookup_connection_value (connection,
+					     MHD_GET_ARGUMENT_KIND,
+					     "wire_method");
+  if (NULL == wire_method)
+  {
+    if (1)
+    {
+      /* temporary work-around until demo is adjusted... */
+      GNUNET_break (0);
+      wire_method = "test";
+      GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+		  "Client needs fixing, see API change for #4943!\n");
+    }
+    else
+      return TMH_RESPONSE_reply_arg_missing (connection,
+					     TALER_EC_PARAMETER_MISSING,
+					     "wire_method");
+  }
+  rctx->wire_method = GNUNET_strdup (wire_method);
 
   instance_str = MHD_lookup_connection_value (connection,
                                               MHD_GET_ARGUMENT_KIND,
@@ -832,11 +996,22 @@ MH_handler_track_transfer (struct TMH_RequestHandler *rh,
   }
 
   /* Check if reply is already in database! */
-  ret = db->find_proof_by_wtid (db->cls,
-                                rctx->uri,
-                                &rctx->wtid,
-                                &proof_cb,
-                                rctx);
+  db->preflight (db->cls);
+  qs = db->find_proof_by_wtid (db->cls,
+			       rctx->url,
+			       &rctx->wtid,
+			       &proof_cb,
+			       rctx);
+  if (0 > qs)
+  {
+    /* Simple select queries should not cause serialization issues */
+    GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
+    /* Always report on hard error as well to enable diagnostics */
+    GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
+    return TMH_RESPONSE_reply_internal_error (connection,
+					      TALER_EC_TRACK_TRANSFER_DB_FETCH_FAILED,
+					      "Fail to query database about proofs");
+  }
   if (0 != rctx->response_code)
   {
     ret = MHD_queue_response (connection,
@@ -857,7 +1032,7 @@ MH_handler_track_transfer (struct TMH_RequestHandler *rh,
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
               "Suspending /track/transfer handling while working with the exchange\n");
   MHD_suspend_connection (connection);
-  rctx->fo = TMH_EXCHANGES_find_exchange (uri,
+  rctx->fo = TMH_EXCHANGES_find_exchange (url,
                                           NULL,
                                           &process_track_transfer_with_exchange,
                                           rctx);
