@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015, 2016 GNUnet e.V. and INRIA
+  Copyright (C) 2014-2017 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Lesser General Public License as published by the Free Software
@@ -13,9 +13,18 @@
   You should have received a copy of the GNU Lesser General Public License along with
   TALER; see the file COPYING.LGPL.  If not, see <http://www.gnu.org/licenses/>
 */
-
+/**
+ * @file taler-merchant-generate-payments.c
+ * @brief tool to use (or test) the taler backend by running some transactions
+ * @author Christian Grothoff
+ * @author Marcello Stanisci
+ *
+ * TODO:
+ * - trigger wirewatch after /admin/add/incoming!
+ */
 #include "platform.h"
 #include <taler/taler_exchange_service.h>
+#include <taler/taler_bank_service.h>
 #include <taler/taler_fakebank_lib.h>
 #include <taler/taler_json_lib.h>
 #include <taler/taler_util.h>
@@ -26,6 +35,11 @@
 #include <gnunet/gnunet_curl_lib.h>
 #include <microhttpd.h>
 
+/**
+ * Number of the account the exchange has at the bank.
+ */
+#define EXCHANGE_ACCOUNT_NO 2
+
 /**
  * The exchange process launched by the generator
  */
@@ -46,6 +60,12 @@ static unsigned int times = 1;
  */
 static unsigned int j = 0;
 
+/**
+ * Indicates whether we use an external exchange.
+ * By default, the generator forks a local exchange.
+ */
+static int remote_bank = 0;
+
 /**
  * Indicates whether we use an external exchange.
  * By default, the generator forks a local exchange.
@@ -60,25 +80,25 @@ static int remote_exchange = 0;
 static int remote_merchant = 0;
 
 /**
- * Exchange URI to withdraw from and deposit to.
+ * Exchange URL to withdraw from and deposit to.
  */
-static char *exchange_uri;
+static char *exchange_url;
 
 /**
  * Base URL of exchange's admin interface.
  */
-static char *exchange_uri_admin;
+static char *exchange_url_admin;
 
 /**
  * Merchant backend to get proposals from and pay.
  */
-static char *merchant_uri;
+static char *merchant_url;
 
 /**
- * Customer's bank URI, communicated at withdrawal time
+ * Customer's bank URL, communicated at withdrawal time
  * to the exchange; must be the same as the exchange's bank.
  */
-static char *bank_uri;
+static char *bank_url;
 
 /**
  * Which merchant instance we use.
@@ -90,6 +110,11 @@ static char *instance;
  */
 static char *currency;
 
+/**
+ * Handle to our fakebank.
+ */
+static struct TALER_FAKEBANK_Handle *fakebank;
+
 /**
  * Task run on timeout.
  */
@@ -120,6 +145,17 @@ static int result;
  */
 static struct GNUNET_DISK_PipeHandle *sigpipe;
 
+/**
+ * Signal handler we overrode.
+ */
+static struct GNUNET_SIGNAL_Context *shc_chld;
+
+/**
+ * Name of the configuration file we are using.
+ */
+static char *cfgfilename;
+
+
 /**
  * State of the interpreter loop.
  */
@@ -168,6 +204,11 @@ enum OpCode
    */
   OC_ADMIN_ADD_INCOMING,
 
+  /**
+   * Run the wirewatcher to check for incoming transactions.
+   */
+  OC_RUN_WIREWATCH,
+
   /**
    * Check status of a reserve.
    */
@@ -251,14 +292,9 @@ struct Command
       struct TALER_DenominationSignature sig;
 
       /**
-       * Set (by the interpreter) to the coin's private key.
+       * Secrets of the planchet. Set by the interpreter.
        */
-      struct TALER_CoinSpendPrivateKeyP coin_priv;
-
-      /**
-       * Blinding key used for the operation.
-       */
-      struct TALER_DenominationBlindingKeyP blinding_key;
+      struct TALER_PlanchetSecretsP ps;
 
       /**
        * Withdraw handle (while operation is running).
@@ -286,14 +322,14 @@ struct Command
       char *amount;
 
       /**
-       * Sender's bank account details (JSON).
+       * Sender's bank account number.
        */
-      const char *sender_details;
+      uint64_t debit_account_no;
 
       /**
-       * Transfer details (JSON)
+       * Receiver's bank account number.
        */
-       const char *transfer_details;
+      uint64_t credit_account_no;
 
       /**
        * Set (by the interpreter) to the reserve's private key
@@ -304,10 +340,29 @@ struct Command
       /**
        * Set to the API's handle during the operation.
        */
-      struct TALER_EXCHANGE_AdminAddIncomingHandle *aih;
+      struct TALER_BANK_AdminAddIncomingHandle *aih;
+
+      /**
+       * Set to bank's identifier for the wire transfer.
+       */
+      uint64_t serial_id;
 
     } admin_add_incoming;
 
+    struct {
+
+      /**
+       * Process for the wirewatcher.
+       */
+      struct GNUNET_OS_Process *wirewatch_proc;
+
+      /**
+       * ID of task called whenever we get a SIGCHILD.
+       */
+      struct GNUNET_SCHEDULER_Task *child_death_task;
+
+    } run_wirewatch;
+
     /**
      * Information for an #OC_PROPOSAL command.
      */
@@ -331,6 +386,11 @@ struct Command
        */
       struct TALER_MERCHANT_ProposalOperation *po;
 
+      /**
+       * Handle to the active GET /proposal operation, or NULL
+       */
+      struct TALER_MERCHANT_ProposalLookupOperation *plo;
+
       /**
        * Full contract in JSON, set by the /contract operation.
        * FIXME: verify in the code that this bit is actually proposal
@@ -464,7 +524,7 @@ next_command (struct InterpreterState *is)
 
 
 /**
- * Callback that works PUT /proposal's output.
+ * Callback that processes GET /proposal's output.
  *
  * @param cls closure
  * @param http_status HTTP response code, 200 indicates success;
@@ -478,18 +538,17 @@ next_command (struct InterpreterState *is)
  * @param h_contract hash of the contract, NULL on error
  */
 static void
-proposal_cb (void *cls,
-             unsigned int http_status,
-	     enum TALER_ErrorCode ec,
-             const json_t *obj,
-             const json_t *contract_terms,
-             const struct TALER_MerchantSignatureP *sig,
-             const struct GNUNET_HashCode *hash)
+proposal_lookup_cb (void *cls,
+                    unsigned int http_status,
+                    const json_t *obj,
+                    const json_t *contract_terms,
+                    const struct TALER_MerchantSignatureP *sig,
+                    const struct GNUNET_HashCode *hash)
 {
   struct InterpreterState *is = cls;
   struct Command *cmd = &is->commands[is->ip];
 
-  cmd->details.proposal.po = NULL;
+  cmd->details.proposal.plo = NULL;
   switch (http_status)
   {
   case MHD_HTTP_OK:
@@ -513,6 +572,59 @@ proposal_cb (void *cls,
   next_command (is);
 }
 
+/**
+ * Callback that works PUT /proposal's output.
+ *
+ * @param cls closure
+ * @param http_status HTTP response code, 200 indicates success;
+ *                    0 if the backend's reply is bogus (fails to follow the protocol)
+ * @param ec taler-specific error code
+ * @param obj the full received JSON reply, or
+ *            error details if the request failed
+ */
+static void
+proposal_cb (void *cls,
+             unsigned int http_status,
+	     enum TALER_ErrorCode ec,
+             const json_t *obj,
+             const char *order_id)
+{
+  struct InterpreterState *is = cls;
+  struct Command *cmd = &is->commands[is->ip];
+
+  cmd->details.proposal.po = NULL;
+  switch (http_status)
+  {
+  case MHD_HTTP_OK: {
+    struct GNUNET_CRYPTO_EddsaPrivateKey *priv;
+    struct GNUNET_CRYPTO_EddsaPublicKey pub;
+
+    priv = GNUNET_CRYPTO_eddsa_key_create ();
+    GNUNET_CRYPTO_eddsa_key_get_public (priv, &pub);
+    GNUNET_free (priv);
+
+    cmd->details.proposal.plo
+        = TALER_MERCHANT_proposal_lookup (ctx,
+                                          merchant_url,
+                                          order_id,
+                                          instance,
+                                          &pub,
+                                          &proposal_lookup_cb,
+                                          is);
+    }
+    break;
+  default:
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "unexpected status code from /proposal: %u. Step %u\n",
+                http_status,
+                is->ip);
+    json_dumpf (obj, stderr, 0);
+    GNUNET_break (0);
+    fail (is);
+    return;
+  }
+}
+
 
 /**
  * Function called with the result of a /pay operation.
@@ -598,24 +710,28 @@ pay_cb (void *cls,
  * @param http_status HTTP response code, #MHD_HTTP_OK (200) for successful status request
  *                    0 if the exchange's reply is bogus (fails to follow the protocol)
  * @param ec taler-specific error code, #TALER_EC_NONE on success
+ * @param serial_id unique ID of for the transfer at the bank
  * @param full_response full response from the exchange (for logging, in case of errors)
  */
 static void
 add_incoming_cb (void *cls,
                  unsigned int http_status,
 		 enum TALER_ErrorCode ec,
+		 uint64_t serial_id,
                  const json_t *full_response)
 {
   struct InterpreterState *is = cls;
   struct Command *cmd = &is->commands[is->ip];
 
   cmd->details.admin_add_incoming.aih = NULL;
+  cmd->details.admin_add_incoming.serial_id = serial_id;
   if (MHD_HTTP_OK != http_status)
   {
     GNUNET_break (0);
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                 "%s",
-                json_dumps (full_response, JSON_INDENT (2)));
+                json_dumps (full_response,
+			    JSON_INDENT (2)));
     fail (is);
     return;
   }
@@ -767,25 +883,47 @@ find_pk (const struct TALER_EXCHANGE_Keys *keys,
  * Allocates and return a string representing a order.
  * In this process, this function gives the order those
  * prices specified by the user. Please NOTE that any amount
- * must be given in the form "XX.YY".
+ * must be given in the form "CUR:XX.YY".
  *
  * @param max_fee merchant's allowed max_fee
  * @param amount total amount for this order
+ * @return JSON string for the order, NULL on errors
  */
 static json_t *
-make_order (char *maxfee,
-            char *total)
+make_order (const char *maxfee,
+            const char *total)
 {
   struct TALER_Amount tmp_amount;
   json_t *total_j;
   json_t *maxfee_j;
   json_t *ret;
   unsigned long long id;
+  struct GNUNET_TIME_Absolute now;
+  char *timestamp;
 
-  TALER_string_to_amount (maxfee, &tmp_amount);
+  if (GNUNET_OK !=
+      TALER_string_to_amount (maxfee,
+                              &tmp_amount))
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
   maxfee_j = TALER_JSON_from_amount (&tmp_amount);
-  TALER_string_to_amount (total, &tmp_amount);
+  GNUNET_assert (NULL != maxfee_j);
+  if (GNUNET_OK !=
+      TALER_string_to_amount (total,
+                              &tmp_amount))
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
   total_j = TALER_JSON_from_amount (&tmp_amount);
+  GNUNET_assert (NULL != total_j);
+  now = GNUNET_TIME_absolute_get ();
+  GNUNET_TIME_round_abs (&now);
+  GNUNET_asprintf (&timestamp,
+                   "/Date(%u)/",
+                   now.abs_value_us / 1000LL / 1000LL);
 
   GNUNET_CRYPTO_random_block (GNUNET_CRYPTO_QUALITY_WEAK,
                               &id,
@@ -793,7 +931,7 @@ make_order (char *maxfee,
   ret = json_pack ("{s:o, s:s, s:s, s:s, s:s, s:o, s:s, s:[{s:s}]}",
                    "max_fee", maxfee_j,
                    "order_id", TALER_b2s (&id, sizeof (id)),
-                   "timestamp", "/Date(42)/",
+                   "timestamp", timestamp,
                    "refund_deadline", "/Date(0)/",
                    "pay_deadline", "/Date(9999999999)/",
                    "amount", total_j,
@@ -806,7 +944,7 @@ make_order (char *maxfee,
 
 
 /**
- * Free amount stringsin interpreter state.
+ * Free amount strings in interpreter state.
  *
  * @param is state to reset
  */
@@ -835,6 +973,8 @@ free_interpreter_amounts (struct InterpreterState *is)
     case OC_ADMIN_ADD_INCOMING:
       GNUNET_free (cmd->details.admin_add_incoming.amount);
       break;
+    case OC_RUN_WIREWATCH:
+      break;
     default:
       GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                   "Shutdown: unknown instruction %d at %u (%s)\n",
@@ -916,10 +1056,26 @@ reset_interpreter (struct InterpreterState *is)
                     "Command %u (%s) did not complete\n",
                     i,
                     cmd->label);
-        TALER_EXCHANGE_admin_add_incoming_cancel (cmd->details.admin_add_incoming.aih);
+        TALER_BANK_admin_add_incoming_cancel (cmd->details.admin_add_incoming.aih);
         cmd->details.admin_add_incoming.aih = NULL;
       }
       break;
+    case OC_RUN_WIREWATCH:
+      if (NULL != cmd->details.run_wirewatch.wirewatch_proc)
+      {
+        GNUNET_break (0 ==
+                      GNUNET_OS_process_kill (cmd->details.run_wirewatch.wirewatch_proc,
+                                              SIGKILL));
+        GNUNET_OS_process_wait (cmd->details.run_wirewatch.wirewatch_proc);
+        GNUNET_OS_process_destroy (cmd->details.run_wirewatch.wirewatch_proc);
+        cmd->details.run_wirewatch.wirewatch_proc = NULL;
+      }
+      if (NULL != cmd->details.run_wirewatch.child_death_task)
+      {
+        GNUNET_SCHEDULER_cancel (cmd->details.run_wirewatch.child_death_task);
+        cmd->details.run_wirewatch.child_death_task = NULL;
+      }
+      break;
     default:
       GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                   "Shutdown: unknown instruction %d at %u (%s)\n",
@@ -931,6 +1087,30 @@ reset_interpreter (struct InterpreterState *is)
 }
 
 
+/**
+ * Task triggered whenever we receive a SIGCHLD (child
+ * process died).
+ *
+ * @param cls closure, NULL if we need to self-restart
+ */
+static void
+maint_child_death (void *cls)
+{
+  struct InterpreterState *is = cls;
+  struct Command *cmd = &is->commands[is->ip];
+  const struct GNUNET_DISK_FileHandle *pr;
+  char c[16];
+
+  cmd->details.run_wirewatch.child_death_task = NULL;
+  pr = GNUNET_DISK_pipe_handle (sigpipe, GNUNET_DISK_PIPE_END_READ);
+  GNUNET_break (0 < GNUNET_DISK_file_read (pr, &c, sizeof (c)));
+  GNUNET_OS_process_wait (cmd->details.run_wirewatch.wirewatch_proc);
+  GNUNET_OS_process_destroy (cmd->details.run_wirewatch.wirewatch_proc);
+  cmd->details.run_wirewatch.wirewatch_proc = NULL;
+  next_command (is);
+}
+
+
 /**
  * Run the main interpreter loop that performs exchange operations.
  *
@@ -943,12 +1123,7 @@ interpreter_run (void *cls)
   struct InterpreterState *is = cls;
   struct Command *cmd = &is->commands[is->ip];
   const struct Command *ref;
-  struct TALER_ReservePublicKeyP reserve_pub;
-  struct TALER_CoinSpendPublicKeyP coin_pub;
   struct TALER_Amount amount;
-  struct GNUNET_TIME_Absolute execution_date;
-  json_t *sender_details;
-  json_t *transfer_details;
 
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
 	      "Interpreter runs command %u/%s(%u)\n",
@@ -969,25 +1144,20 @@ interpreter_run (void *cls)
   switch (cmd->oc)
   {
     case OC_END:
-
       j++;
-
       if (j < times)
       {
         reset_interpreter (is);
         is->ip = 0;
         GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
                     "Rewinding the interpreter.\n");
-
         GNUNET_SCHEDULER_add_now (&interpreter_run,
                                   is);
         return;
       }
       result = GNUNET_OK;
       GNUNET_SCHEDULER_shutdown ();
-
       return;
-
     case OC_PAY:
       {
         struct TALER_MERCHANT_PayCoin pc;
@@ -1042,17 +1212,19 @@ interpreter_run (void *cls)
 
         {
           const struct Command *coin_ref;
+
           memset (&pc, 0, sizeof (pc));
           coin_ref = find_command (is,
                                    cmd->details.pay.coin_ref);
-          GNUNET_assert (NULL != ref);
+          GNUNET_assert (NULL != coin_ref);
           switch (coin_ref->oc)
           {
           case OC_WITHDRAW_SIGN:
-            pc.coin_priv = coin_ref->details.reserve_withdraw.coin_priv;
+            pc.coin_priv = coin_ref->details.reserve_withdraw.ps.coin_priv;
             pc.denom_pub = coin_ref->details.reserve_withdraw.pk->key;
             pc.denom_sig = coin_ref->details.reserve_withdraw.sig;
             pc.denom_value = coin_ref->details.reserve_withdraw.pk->value;
+	    pc.exchange_url = exchange_url;
             break;
           default:
             GNUNET_assert (0);
@@ -1084,7 +1256,7 @@ interpreter_run (void *cls)
         }
         cmd->details.pay.ph
           = TALER_MERCHANT_pay_wallet (ctx,
-                                       merchant_uri,
+                                       merchant_url,
                                        instance,
                                        &ref->details.proposal.hash,
                                        &total_amount,
@@ -1095,7 +1267,6 @@ interpreter_run (void *cls)
                                        refund_deadline,
                                        pay_deadline,
                                        &h_wire,
-                                       exchange_uri,
                                        order_id,
                                        1 /* num_coins */,
                                        &pc /* coins */,
@@ -1138,7 +1309,7 @@ interpreter_run (void *cls)
         json_decref (merchant_obj);
         cmd->details.proposal.po
           = TALER_MERCHANT_order_put (ctx,
-                                      merchant_uri,
+                                      merchant_url,
                                       order,
                                       &proposal_cb,
                                       is);
@@ -1153,92 +1324,102 @@ interpreter_run (void *cls)
       }
 
     case OC_ADMIN_ADD_INCOMING:
-      if (NULL !=
-          cmd->details.admin_add_incoming.reserve_reference)
       {
-        ref = find_command (is,
-                            cmd->details.admin_add_incoming.reserve_reference);
-        GNUNET_assert (NULL != ref);
-        GNUNET_assert (OC_ADMIN_ADD_INCOMING == ref->oc);
-        cmd->details.admin_add_incoming.reserve_priv
-          = ref->details.admin_add_incoming.reserve_priv;
-      }
-      else
-      {
-        struct GNUNET_CRYPTO_EddsaPrivateKey *priv;
+	char *subject;
+	struct TALER_BANK_AuthenticationData auth;
+	struct TALER_ReservePublicKeyP reserve_pub;
 
-        priv = GNUNET_CRYPTO_eddsa_key_create ();
-        cmd->details.admin_add_incoming.reserve_priv.eddsa_priv = *priv;
-        GNUNET_free (priv);
-      }
-      GNUNET_CRYPTO_eddsa_key_get_public (&cmd->details.admin_add_incoming.reserve_priv.eddsa_priv,
-                                          &reserve_pub.eddsa_pub);
-      if (GNUNET_OK !=
-          TALER_string_to_amount (cmd->details.admin_add_incoming.amount,
-                                  &amount))
-      {
-        GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                    "Failed to parse amount `%s' at %u\n",
-                    cmd->details.admin_add_incoming.amount,
-                    is->ip);
-        fail (is);
-        return;
-      }
+	if (NULL !=
+	    cmd->details.admin_add_incoming.reserve_reference)
+	{
+	  ref = find_command (is,
+			      cmd->details.admin_add_incoming.reserve_reference);
+	  GNUNET_assert (NULL != ref);
+	  GNUNET_assert (OC_ADMIN_ADD_INCOMING == ref->oc);
+	  cmd->details.admin_add_incoming.reserve_priv
+	    = ref->details.admin_add_incoming.reserve_priv;
+	}
+	else
+	{
+	  struct GNUNET_CRYPTO_EddsaPrivateKey *priv;
 
-      execution_date = GNUNET_TIME_absolute_get ();
-      GNUNET_TIME_round_abs (&execution_date);
-      sender_details = json_loads (cmd->details.admin_add_incoming.sender_details,
-                                   JSON_REJECT_DUPLICATES,
-                                   NULL);
-      if (NULL == sender_details)
-      {
-        GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                    "Failed to parse sender details `%s' at %u\n",
-                    cmd->details.admin_add_incoming.sender_details,
-                    is->ip);
-        fail (is);
-        return;
-      }
-      json_object_set_new (sender_details,
-                           "bank_uri",
-                           json_string (bank_uri));
+	  priv = GNUNET_CRYPTO_eddsa_key_create ();
+	  cmd->details.admin_add_incoming.reserve_priv.eddsa_priv = *priv;
+	  GNUNET_free (priv);
+	}
+	GNUNET_CRYPTO_eddsa_key_get_public (&cmd->details.admin_add_incoming.reserve_priv.eddsa_priv,
+					    &reserve_pub.eddsa_pub);
+	if (GNUNET_OK !=
+	    TALER_string_to_amount (cmd->details.admin_add_incoming.amount,
+				    &amount))
+	{
+	  GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+		      "Failed to parse amount `%s' at %u\n",
+		      cmd->details.admin_add_incoming.amount,
+		      is->ip);
+	  fail (is);
+	  return;
+	}
 
-      transfer_details = json_loads (cmd->details.admin_add_incoming.transfer_details,
-                                     JSON_REJECT_DUPLICATES,
-                                     NULL);
-
-      if (NULL == transfer_details)
-      {
-        GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                    "Failed to parse transfer details `%s' at %u\n",
-                    cmd->details.admin_add_incoming.transfer_details,
-                    is->ip);
-        fail (is);
-        return;
+	subject
+	  = GNUNET_STRINGS_data_to_string_alloc (&reserve_pub,
+						 sizeof (reserve_pub));
+	auth.method = TALER_BANK_AUTH_BASIC;
+	/* TODO: obtain authentication details from configuration */
+	auth.details.basic.username = "admin";
+	auth.details.basic.password = "x";
+	cmd->details.admin_add_incoming.aih
+	  = TALER_BANK_admin_add_incoming (ctx,
+					   bank_url,
+					   &auth,
+					   exchange_url,
+					   subject,
+					   &amount,
+					   cmd->details.admin_add_incoming.debit_account_no,
+					   cmd->details.admin_add_incoming.credit_account_no,
+					   &add_incoming_cb,
+					   is);
+	GNUNET_free (subject);
+	if (NULL == cmd->details.admin_add_incoming.aih)
+	{
+	  GNUNET_break (0);
+	  fail (is);
+	  return;
+	}
+	return;
       }
-      cmd->details.admin_add_incoming.aih
-        = TALER_EXCHANGE_admin_add_incoming (exchange,
-                                             exchange_uri_admin,
-                                             &reserve_pub,
-                                             &amount,
-                                             execution_date,
-                                             sender_details,
-                                             transfer_details,
-                                             &add_incoming_cb,
-                                             is);
-      json_decref (sender_details);
-      json_decref (transfer_details);
-      if (NULL == cmd->details.admin_add_incoming.aih)
+    case OC_RUN_WIREWATCH:
       {
-        GNUNET_break (0);
-        fail (is);
-        return;
-      }
-      return;
+	const struct GNUNET_DISK_FileHandle *pr;
 
+	cmd->details.run_wirewatch.wirewatch_proc
+	  = GNUNET_OS_start_process (GNUNET_NO,
+				     GNUNET_OS_INHERIT_STD_ALL,
+				     NULL, NULL, NULL,
+				     "taler-exchange-wirewatch",
+				     "taler-exchange-wirewatch",
+				     "-c", cfgfilename,
+				     "-t", "test", /* use Taler's bank/fakebank */
+				     "-T", /* exit when done */
+				     NULL);
+	if (NULL == cmd->details.run_wirewatch.wirewatch_proc)
+	{
+	  GNUNET_break (0);
+	  fail (is);
+	  return;
+	}
+	pr = GNUNET_DISK_pipe_handle (sigpipe,
+				      GNUNET_DISK_PIPE_END_READ);
+	cmd->details.run_wirewatch.child_death_task
+	  = GNUNET_SCHEDULER_add_read_file (GNUNET_TIME_UNIT_FOREVER_REL,
+					    pr,
+					    &maint_child_death,
+                                            is);
+	return;
+      }
     case OC_WITHDRAW_SIGN:
       GNUNET_assert (NULL !=
-                     cmd->details.reserve_withdraw.reserve_reference);
+		     cmd->details.reserve_withdraw.reserve_reference);
       ref = find_command (is,
                           cmd->details.reserve_withdraw.reserve_reference);
       GNUNET_assert (NULL != ref);
@@ -1268,26 +1449,12 @@ interpreter_run (void *cls)
         return;
       }
 
-      /* create coin's private key */
-      {
-        struct GNUNET_CRYPTO_EddsaPrivateKey *priv;
-
-        priv = GNUNET_CRYPTO_eddsa_key_create ();
-        cmd->details.reserve_withdraw.coin_priv.eddsa_priv = *priv;
-        GNUNET_free (priv);
-      }
-      GNUNET_CRYPTO_eddsa_key_get_public (&cmd->details.reserve_withdraw.coin_priv.eddsa_priv,
-                                          &coin_pub.eddsa_pub);
-      GNUNET_CRYPTO_random_block (GNUNET_CRYPTO_QUALITY_WEAK,
-                                  &cmd->details.reserve_withdraw.blinding_key,
-                                  sizeof (cmd->details.reserve_withdraw.blinding_key));
-
+      TALER_planchet_setup_random (&cmd->details.reserve_withdraw.ps);
       cmd->details.reserve_withdraw.wsh
         = TALER_EXCHANGE_reserve_withdraw (exchange,
                                            cmd->details.reserve_withdraw.pk,
                                            &ref->details.admin_add_incoming.reserve_priv,
-                                           &cmd->details.reserve_withdraw.coin_priv,
-                                           &cmd->details.reserve_withdraw.blinding_key,
+                                           &cmd->details.reserve_withdraw.ps,
                                            &reserve_withdraw_cb,
                                            is);
       if (NULL == cmd->details.reserve_withdraw.wsh)
@@ -1315,10 +1482,12 @@ interpreter_run (void *cls)
  *
  * @param cls closure
  * @param keys information about keys of the exchange
+ * @param compat version compatibility data
  */
 static void
 cert_cb (void *cls,
-         const struct TALER_EXCHANGE_Keys *keys)
+         const struct TALER_EXCHANGE_Keys *keys,
+	 enum TALER_EXCHANGE_VersionCompatibility compat)
 {
   struct InterpreterState *is = cls;
 
@@ -1403,6 +1572,11 @@ do_shutdown (void *cls)
     GNUNET_CURL_gnunet_rc_destroy (rc);
     rc = NULL;
   }
+  if (NULL != fakebank)
+  {
+    TALER_FAKEBANK_stop (fakebank);
+    fakebank = NULL;
+  }
 }
 
 
@@ -1417,12 +1591,15 @@ do_shutdown (void *cls)
  * formatted string.
  */
 static char *
-concat_amount (char *currency, char *rpart)
+concat_amount (const char *currency,
+	       const char *rpart)
 {
   char *str;
 
-  GNUNET_asprintf (&str, "%s:%s",
-                   currency, rpart);
+  GNUNET_asprintf (&str,
+		   "%s:%s",
+                   currency,
+		   rpart);
   return str;
 }
 
@@ -1440,25 +1617,27 @@ run_test ()
     { .oc = OC_ADMIN_ADD_INCOMING,
       .label = "create-reserve-1",
       .expected_response_code = MHD_HTTP_OK,
-      .details.admin_add_incoming.sender_details = "{ \"type\":\"test\", \"account_number\":62, \"uuid\":1 }",
-      .details.admin_add_incoming.transfer_details = "{ \"uuid\": 1}",
+      .details.admin_add_incoming.debit_account_no = 62,
+      .details.admin_add_incoming.credit_account_no = EXCHANGE_ACCOUNT_NO,
       .details.admin_add_incoming.amount = concat_amount (currency, "5.01") },
 
     /* Fill reserve with EUR:5.01, as withdraw fee is 1 ct per config */
     { .oc = OC_ADMIN_ADD_INCOMING,
       .label = "create-reserve-2",
       .expected_response_code = MHD_HTTP_OK,
-      .details.admin_add_incoming.sender_details = "{ \"type\":\"test\", \"account_number\":62, \"uuid\":1 }",
-      .details.admin_add_incoming.transfer_details = "{ \"uuid\": 1}",
+      .details.admin_add_incoming.debit_account_no = 62,
+      .details.admin_add_incoming.credit_account_no = EXCHANGE_ACCOUNT_NO,
       .details.admin_add_incoming.amount = concat_amount (currency, "5.01") },
     /* Fill reserve with EUR:5.01, as withdraw fee is 1 ct per config */
     { .oc = OC_ADMIN_ADD_INCOMING,
       .label = "create-reserve-3",
       .expected_response_code = MHD_HTTP_OK,
-      .details.admin_add_incoming.sender_details = "{ \"type\":\"test\", \"account_number\":62, \"uuid\":1 }",
-      .details.admin_add_incoming.transfer_details = "{ \"uuid\": 1}",
+      .details.admin_add_incoming.debit_account_no = 62,
+      .details.admin_add_incoming.credit_account_no = EXCHANGE_ACCOUNT_NO,
       .details.admin_add_incoming.amount = concat_amount (currency, "5.01") },
-
+    { .oc = OC_RUN_WIREWATCH,
+      .label = "run-wirewatch-1"
+    },
     /* Withdraw a 5 EUR coin, at fee of 1 ct */
     { .oc = OC_WITHDRAW_SIGN,
       .label = "withdraw-coin-1",
@@ -1539,7 +1718,7 @@ run_test ()
   GNUNET_assert (NULL != ctx);
   rc = GNUNET_CURL_gnunet_rc_create (ctx);
   exchange = TALER_EXCHANGE_connect (ctx,
-                                     exchange_uri,
+                                     exchange_url,
                                      &cert_cb,
                                      is,
                                      TALER_EXCHANGE_OPTION_END);
@@ -1567,10 +1746,23 @@ run (void *cls,
   unsigned int cnt;
   char *wget_cmd;
 
-  if (GNUNET_SYSERR == GNUNET_CONFIGURATION_get_value_string (config,
-                                                              "payments-generator",
-                                                              "exchange",
-                                                              &exchange_uri))
+  cfgfilename = GNUNET_strdup (cfgfile);
+  if (! remote_bank)
+  {
+    /* TODO: do not hard-code port, find in cfg */
+    fakebank = TALER_FAKEBANK_start (8888);
+    if (NULL == fakebank)
+    {
+      fprintf (stderr,
+	       "Failed to launch fakebank\n");
+      GNUNET_SCHEDULER_shutdown ();
+    }
+  }
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_string (config,
+					     "payments-generator",
+					     "exchange",
+					     &exchange_url))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "payments-generator",
@@ -1578,10 +1770,11 @@ run (void *cls,
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
-  if (GNUNET_SYSERR == GNUNET_CONFIGURATION_get_value_string (config,
-                                                              "payments-generator",
-                                                              "exchange_admin",
-                                                              &exchange_uri_admin))
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_string (config,
+					     "payments-generator",
+					     "exchange_admin",
+					     &exchange_url_admin))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "payments-generator",
@@ -1590,10 +1783,11 @@ run (void *cls,
     return;
   }
 
-  if (GNUNET_SYSERR == GNUNET_CONFIGURATION_get_value_string (config,
-                                                              "payments-generator",
-                                                              "merchant",
-                                                              &merchant_uri))
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_string (config,
+					     "payments-generator",
+					     "merchant",
+					     &merchant_url))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "payments-generator",
@@ -1602,12 +1796,12 @@ run (void *cls,
     return;
   }
 
-  if (GNUNET_SYSERR == GNUNET_CONFIGURATION_get_value_string (config,
-                                                              "payments-generator",
-                                                              "bank",
-                                                              &bank_uri))
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_string (config,
+					     "payments-generator",
+					     "bank",
+					     &bank_url))
   {
-
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "payments-generator",
                                "bank");
@@ -1615,10 +1809,11 @@ run (void *cls,
     return;
   }
 
-  if (GNUNET_SYSERR == GNUNET_CONFIGURATION_get_value_string (config,
-                                                              "payments-generator",
-                                                              "instance",
-                                                              &instance))
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_string (config,
+					     "payments-generator",
+					     "instance",
+					     &instance))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "payments-generator",
@@ -1627,12 +1822,12 @@ run (void *cls,
     return;
   }
 
-  if (GNUNET_SYSERR == GNUNET_CONFIGURATION_get_value_string (config,
-                                                              "payments-generator",
-                                                              "currency",
-                                                              &currency))
+  if (GNUNET_SYSERR ==
+      GNUNET_CONFIGURATION_get_value_string (config,
+					     "payments-generator",
+					     "currency",
+					     &currency))
   {
-
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "payments-generator",
                                "currency");
@@ -1640,7 +1835,7 @@ run (void *cls,
     return;
   }
 
-  if (!remote_exchange)
+  if (! remote_exchange)
   {
     exchanged = GNUNET_OS_start_process (GNUNET_NO,
                                          GNUNET_OS_INHERIT_STD_ALL,
@@ -1660,7 +1855,9 @@ run (void *cls,
              "Waiting for taler-exchange-httpd to be ready\n");
     cnt = 0;
 
-    GNUNET_asprintf (&wget_cmd, "wget -q -t 1 -T 1 %skeys -o /dev/null -O /dev/null", exchange_uri);
+    GNUNET_asprintf (&wget_cmd,
+		     "wget -q -t 1 -T 1 %skeys -o /dev/null -O /dev/null",
+		     exchange_url);
 
     do
       {
@@ -1702,7 +1899,6 @@ run (void *cls,
                               SIGKILL);
       GNUNET_OS_process_wait (exchanged);
       GNUNET_OS_process_destroy (exchanged);
-
       GNUNET_SCHEDULER_shutdown ();
       return;
     }
@@ -1710,7 +1906,9 @@ run (void *cls,
     fprintf (stderr,
              "Waiting for taler-merchant-httpd to be ready\n");
     cnt = 0;
-    GNUNET_asprintf (&wget_cmd, "wget -q -t 1 -T 1 %s -o /dev/null -O /dev/null", merchant_uri);
+    GNUNET_asprintf (&wget_cmd,
+                     "wget -q -t 1 -T 1 %s -o /dev/null -O /dev/null",
+                     merchant_url);
 
     do
       {
@@ -1729,7 +1927,6 @@ run (void *cls,
                                   SIGKILL);
           GNUNET_OS_process_wait (exchanged);
           GNUNET_OS_process_destroy (exchanged);
-
           GNUNET_SCHEDULER_shutdown ();
           return;
         }
@@ -1738,9 +1935,18 @@ run (void *cls,
     fprintf (stderr, "\n");
     GNUNET_free (wget_cmd);
   }
+
+  shc_chld = GNUNET_SIGNAL_handler_install (GNUNET_SIGCHLD,
+                                            &sighandler_child_death);
+
+  /* timeout, given 60s + 5s per command, which should be more
+     than enough */
   timeout_task
-    = GNUNET_SCHEDULER_add_delayed (GNUNET_TIME_relative_multiply
-                                    (GNUNET_TIME_UNIT_SECONDS, 150),
+    = GNUNET_SCHEDULER_add_delayed (GNUNET_TIME_relative_add
+				    (GNUNET_TIME_UNIT_MINUTES,
+				     GNUNET_TIME_relative_multiply
+				     (GNUNET_TIME_UNIT_SECONDS,
+				      5 * times)),
                                     &do_timeout, NULL);
   run_test ();
 }
@@ -1750,13 +1956,16 @@ int
 main (int argc,
       char *argv[])
 {
-  struct GNUNET_SIGNAL_Context *shc_chld;
   struct GNUNET_GETOPT_CommandLineOption options[] = {
     GNUNET_GETOPT_option_uint ('n',
                                "times",
                                "TIMES",
                                "How many times the commands should be run.",
                                &times),
+    GNUNET_GETOPT_option_flag ('b',
+                               "remote-bank",
+                               "Do not start fakebank",
+                               &remote_bank),
     GNUNET_GETOPT_option_flag ('e',
                                "remote-exchange",
                                "Do not fork any exchange",
@@ -1774,11 +1983,11 @@ main (int argc,
                     "DEBUG",
                     NULL);
   result = GNUNET_SYSERR;
+
   sigpipe = GNUNET_DISK_pipe (GNUNET_NO, GNUNET_NO,
                               GNUNET_NO, GNUNET_NO);
   GNUNET_assert (NULL != sigpipe);
-  shc_chld = GNUNET_SIGNAL_handler_install (GNUNET_SIGCHLD,
-                                            &sighandler_child_death);
+
   if (GNUNET_OK !=
       GNUNET_PROGRAM_run (argc, argv,
                           "taler-merchant-generate-payments",
@@ -1787,9 +1996,11 @@ main (int argc,
                           &run, NULL))
     return 77;
 
-
-  GNUNET_SIGNAL_handler_uninstall (shc_chld);
-  shc_chld = NULL;
+  if (NULL != shc_chld)
+  {
+    GNUNET_SIGNAL_handler_uninstall (shc_chld);
+    shc_chld = NULL;
+  }
   GNUNET_DISK_pipe_close (sigpipe);
   if (!remote_merchant && NULL != merchantd)
   {
